"""
Validate ODC dataset documents
"""

import collections
import enum
import math
import multiprocessing
import os
import sys
from datetime import datetime
from functools import partial
from pathlib import Path
from textwrap import indent
from typing import (
    Counter,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from urllib.parse import urljoin, urlparse
from urllib.request import urlopen

import attr
import cattr
import ciso8601
import click
import numpy as np
import rasterio
import toolz
from attr import Factory, define, field, frozen
from boltons.iterutils import get_path
from click import echo, secho, style
from datacube import Datacube
from datacube.index.eo3 import prep_eo3
from datacube.utils import InvalidDocException, changes, is_url, read_documents
from datacube.utils.documents import load_documents
from rasterio import DatasetReader
from rasterio.crs import CRS
from rasterio.errors import CRSError
from shapely.validation import explain_validity

from eodatasets3 import model, serialise, utils
from eodatasets3.model import DatasetDoc
from eodatasets3.ui import bool_style, is_absolute, uri_resolve
from eodatasets3.utils import EO3_SCHEMA, default_utc

DEFAULT_NULLABLE_FIELDS = ("label",)
DEFAULT_OPTIONAL_FIELDS = (
    # Older product do not have this field at all, and when not specified it is considered stable.
    "dataset_maturity",
)


class Level(enum.Enum):
    info = 1
    warning = 2
    error = 3


class DocKind(enum.Enum):
    # EO3 datacube dataset.
    dataset = 1
    # Datacube product
    product = 2
    # Datacube Metadata Type
    metadata_type = 3
    # Stac Item
    stac_item = 4
    # Legacy datacube ("eo1") dataset
    legacy_dataset = 5
    # Legacy product config for ingester
    ingestion_config = 6


# What kind of document each suffix represents.
# (full suffix will also have a doc type: .yaml, .json, .yaml.gz etc)
# Example:  "my-test-dataset.odc-metadata.yaml"
SUFFIX_KINDS = {
    ".odc-metadata": DocKind.dataset,
    ".odc-product": DocKind.product,
    ".odc-type": DocKind.metadata_type,
}
# Inverse of above
DOC_TYPE_SUFFIXES = {v: k for k, v in SUFFIX_KINDS.items()}


def filename_doc_kind(path: Union[str, Path]) -> Optional["DocKind"]:
    """
    Get the expected file type for the given filename.

    Returns None if it does not follow any naming conventions.

    >>> filename_doc_kind('LC8_2014.odc-metadata.yaml').name
    'dataset'
    >>> filename_doc_kind('/tmp/something/water_bodies.odc-metadata.yaml.gz').name
    'dataset'
    >>> filename_doc_kind(Path('/tmp/something/ls8_fc.odc-product.yaml')).name
    'product'
    >>> filename_doc_kind(Path('/tmp/something/ls8_wo.odc-product.json.gz')).name
    'product'
    >>> filename_doc_kind(Path('/tmp/something/eo3_gqa.odc-type.yaml')).name
    'metadata_type'
    >>> filename_doc_kind(Path('/tmp/something/some_other_file.yaml'))
    """

    for suffix in reversed(Path(path).suffixes):
        suffix = suffix.lower()
        if suffix in SUFFIX_KINDS:
            return SUFFIX_KINDS[suffix]

    return None


def guess_kind_from_contents(doc: Dict):
    """
    What sort of document do the contents look like?
    """
    if "$schema" in doc and doc["$schema"] == EO3_SCHEMA:
        return DocKind.dataset
    if "metadata_type" in doc:
        if "source_type" in doc:
            return DocKind.ingestion_config
        return DocKind.product
    if ("dataset" in doc) and ("search_fields" in doc["dataset"]):
        return DocKind.metadata_type
    if "id" in doc:
        if ("lineage" in doc) and ("platform" in doc):
            return DocKind.legacy_dataset

        if ("properties" in doc) and ("datetime" in doc["properties"]):
            return DocKind.stac_item

    return None


@frozen
class ValidationMessage:
    level: Level
    code: str
    reason: str
    hint: str = None
    #: What was assumed when validating this? Eg: dict(product='ls7_nbar', metadata_type='eo3')
    context: dict = None

    def __str__(self) -> str:
        hint = ""
        if self.hint:
            hint = f" (Hint: {self.hint})"
        return f"{self.code}: {self.reason}{hint}"


def _info(code: str, reason: str, hint: str = None):
    return ValidationMessage(Level.info, code, reason, hint=hint)


def _warning(code: str, reason: str, hint: str = None):
    return ValidationMessage(Level.warning, code, reason, hint=hint)


def _error(code: str, reason: str, hint: str = None):
    return ValidationMessage(Level.error, code, reason, hint=hint)


ValidationMessages = Generator[ValidationMessage, None, None]


@frozen(init=True)
class ValidationExpectations:
    """
    What expectations do we have when validating this dataset?
    """

    #: Allow these extra measurement names to be included in the dataset.
    #: (ODC allows unlisted measurement names, but it's usually a mistake)
    allow_extra_measurements: Sequence[str] = ()

    #: Do we expect full geometry information in every dataset?
    #: (It's optional in ODC, but often a mistake to miss it)
    require_geometry: bool = True

    #: Are any of the configured fields nullable?
    allow_nullable_fields: Sequence[str] = field(
        default=Factory(lambda: DEFAULT_NULLABLE_FIELDS)
    )
    #: Can any of the fields be completely omitted from the document?
    allow_missing_fields: Sequence[str] = field(
        default=Factory(lambda: DEFAULT_OPTIONAL_FIELDS)
    )

    def with_document_overrides(self, doc: Dict):
        """
        Return an instance with any overrides from the given document.
        """
        if "default_allowances" not in doc:
            return self

        overridden_values = {**attr.asdict(self), **doc["default_allowances"]}
        # Merge, don't replace, these lists.
        overridden_values["allow_nullable_fields"] = list(
            {*overridden_values["allow_nullable_fields"], *self.allow_nullable_fields}
        )
        overridden_values["allow_missing_fields"] = list(
            {*overridden_values["allow_missing_fields"], *self.allow_missing_fields}
        )
        overridden_values["allow_extra_measurements"] = list(
            {
                *overridden_values["allow_extra_measurements"],
                *self.allow_extra_measurements,
            }
        )
        return cattr.structure(overridden_values, self.__class__)


def validate_dataset(
    doc: Dict,
    product_definition: Optional[Dict] = None,
    metadata_type_definition: Optional[Dict] = None,
    thorough: bool = False,
    readable_location: Union[str, Path] = None,
    expect: ValidationExpectations = None,
) -> ValidationMessages:
    """
    Validate a dataset document, optionally against the given product.

    By default this will only look at the metadata, run with thorough=True to
    open the data files too.

    :param product_definition: Optionally check that the dataset matches this product definition.
    :param thorough: Open the imagery too, to check that data types etc match.
    :param readable_location: Dataset location to use, if not the metadata path.
    :param expect: Where can we be lenient in validation?
    """
    validation_context = {}
    expect = expect or ValidationExpectations()
    if metadata_type_definition is not None:
        expect = expect.with_document_overrides(metadata_type_definition)
        validation_context["type"] = metadata_type_definition["name"]
    if product_definition is not None:
        expect = expect.with_document_overrides(product_definition)
        validation_context["product"] = product_definition["name"]

    # noinspection PyShadowingNames
    def _info(code: str, reason: str, hint: str = None):
        return ValidationMessage(
            Level.info, code, reason, hint=hint, context=validation_context
        )

    # noinspection PyShadowingNames
    def _warning(code: str, reason: str, hint: str = None):
        return ValidationMessage(
            Level.warning, code, reason, hint=hint, context=validation_context
        )

    # noinspection PyShadowingNames
    def _error(code: str, reason: str, hint: str = None):
        return ValidationMessage(
            Level.error, code, reason, hint=hint, context=validation_context
        )

    schema = doc.get("$schema")
    if schema is None:
        yield _error(
            "no_schema",
            f"No $schema field. "
            f"You probably want an ODC dataset schema {model.ODC_DATASET_SCHEMA_URL!r}",
        )
        return
    if schema != model.ODC_DATASET_SCHEMA_URL:
        yield _error(
            "unknown_doc_type",
            f"Unknown doc schema {schema!r}. Only ODC datasets are supported ({model.ODC_DATASET_SCHEMA_URL!r})",
        )
        return

    has_doc_errors = False
    for error in serialise.DATASET_SCHEMA.iter_errors(doc):
        has_doc_errors = True
        displayable_path = ".".join(error.absolute_path)

        hint = None
        if displayable_path == "crs" and "not of type" in error.message:
            hint = "epsg codes should be prefixed with 'epsg:1234'"

        context = f"({displayable_path}) " if displayable_path else ""
        yield _error("structure", f"{context}{error.message} ", hint=hint)

    if has_doc_errors:
        return

    dataset = serialise.from_doc(doc, skip_validation=True)

    if not dataset.product.href:
        _info("product_href", "A url (href) is recommended for products")

    yield from _validate_geo(dataset, expect_geometry=expect.require_geometry)

    # Note that a dataset may have no measurements (eg. telemetry data).
    # (TODO: a stricter mode for when we know we should have geo and measurement info)
    if dataset.measurements:
        for name, measurement in dataset.measurements.items():
            grid_name = measurement.grid
            if grid_name != "default" or dataset.grids:
                if grid_name not in dataset.grids:
                    yield _error(
                        "invalid_grid_ref",
                        f"Measurement {name!r} refers to unknown grid {grid_name!r}",
                    )

            if is_absolute(measurement.path):
                yield _warning(
                    "absolute_path",
                    f"measurement {name!r} has an absolute path: {measurement.path!r}",
                )

    yield from _validate_stac_properties(dataset)

    required_measurements: Dict[str, ExpectedMeasurement] = {}
    if product_definition is not None:
        required_measurements.update(
            {
                m.name: m
                for m in map(
                    ExpectedMeasurement.from_definition,
                    product_definition.get("measurements") or (),
                )
            }
        )

        product_name = product_definition.get("name")
        if product_name != dataset.product.name:
            # This is only informational as it's possible products may be indexed with finer-grained
            # categories than the original datasets: eg. a separate "nrt" product, or test product.
            yield _info(
                "product_mismatch",
                f"Dataset product name {dataset.product.name!r} "
                f"does not match the given product ({product_name!r}",
            )

        for name in required_measurements:
            if name not in dataset.measurements.keys():
                yield _error(
                    "missing_measurement",
                    f"Product {product_name} expects a measurement {name!r})",
                )
        measurements_not_in_product = set(dataset.measurements.keys()).difference(
            {m["name"] for m in product_definition.get("measurements") or ()}
        )
        # Remove the measurements that are allowed to be extra.
        measurements_not_in_product.difference_update(
            expect.allow_extra_measurements or set()
        )

        if measurements_not_in_product:
            things = ", ".join(sorted(measurements_not_in_product))
            yield _warning(
                "extra_measurements",
                f"Dataset has measurements not present in product definition for {product_name!r}: {things}",
                hint="This may be valid, as it's allowed by ODC. Set `expect_extra_measurements` to mute this.",
            )

    if metadata_type_definition:
        # Datacube does certain transforms on an eo3 doc before storage.
        # We need to do the same, as the fields will be read from the storage.
        prepared_doc = prep_eo3(doc)

        all_nullable_fields = tuple(expect.allow_nullable_fields) + tuple(
            expect.allow_missing_fields
        )
        for field_name, offsets in _get_field_offsets(
            metadata_type=metadata_type_definition
        ):
            if (
                # If a field is required...
                (field_name not in expect.allow_missing_fields)
                and
                # ... and none of its offsets are in the document
                not any(_has_offset(prepared_doc, offset) for offset in offsets)
            ):
                # ... warn them.
                product_name = (
                    product_definition.get("name")
                    if product_definition
                    else dataset.product.name
                )
                readable_offsets = " or ".join("->".join(offset) for offset in offsets)
                yield _warning(
                    "missing_field",
                    f"Dataset is missing field {field_name!r} "
                    f"for type {metadata_type_definition['name']!r}",
                    hint=f"Expected at {readable_offsets}",
                )
                continue

            if field_name not in all_nullable_fields:
                value = None
                for offset in offsets:
                    value = toolz.get_in(offset, prepared_doc)
                if value is None:
                    yield _info(
                        "null_field",
                        f"Value is null for configured field {field_name!r}",
                    )

    dataset_location = dataset.locations[0] if dataset.locations else readable_location

    # If we have a location:
    # For each measurement, try to load it.
    # If loadable:
    if thorough:
        for name, measurement in dataset.measurements.items():
            full_path = uri_resolve(dataset_location, measurement.path)
            expected_measurement = required_measurements.get(name)

            band = measurement.band or 1
            with rasterio.open(full_path) as ds:
                ds: DatasetReader

                if band not in ds.indexes:
                    yield _error(
                        "incorrect_band",
                        f"Measurement {name!r} file contains no rio index {band!r}.",
                        hint=f"contains indexes {ds.indexes!r}",
                    )
                    continue

                if not expected_measurement:
                    # The measurement is not in the product definition
                    #
                    # This is only informational because a product doesn't have to define all
                    # measurements that the datasets contain.
                    #
                    # This is historically because dataset documents reflect the measurements that
                    # are stored on disk, which can differ. But products define the set of measurments
                    # that are mandatory in every dataset.
                    #
                    # (datasets differ when, for example, sensors go offline, or when there's on-disk
                    #  measurements like panchromatic that GA doesn't want in their product definitions)
                    if required_measurements:
                        yield _info(
                            "unspecified_measurement",
                            f"Measurement {name} is not in the product",
                        )
                else:
                    expected_dtype = expected_measurement.dtype
                    band_dtype = ds.dtypes[band - 1]
                    # TODO: NaN handling
                    if expected_dtype != band_dtype:
                        yield _error(
                            "different_dtype",
                            f"{name} dtype: "
                            f"product {expected_dtype!r} != dataset {band_dtype!r}",
                        )

                    ds_nodata = ds.nodatavals[band - 1]

                    # If the dataset is missing 'nodata', we can allow anything in product 'nodata'.
                    # (In ODC, nodata might be a fill value for loading data.)
                    if ds_nodata is None:
                        continue

                    # Otherwise check that nodata matches.
                    expected_nodata = expected_measurement.nodata
                    if expected_nodata != ds_nodata and not (
                        _is_nan(expected_nodata) and _is_nan(ds_nodata)
                    ):
                        yield _error(
                            "different_nodata",
                            f"{name} nodata: "
                            f"product {expected_nodata !r} != dataset {ds_nodata !r}",
                        )


def _has_offset(doc: Dict, offset: List[str]) -> bool:
    """
    Is the given offset present in the document?
    """
    for key in offset:
        if key not in doc:
            return False
        doc = doc[key]
    return True


def validate_product(doc: Dict) -> ValidationMessages:
    """
    Check for common product mistakes
    """

    # Validate it against ODC's product schema.
    has_doc_errors = False
    for error in serialise.PRODUCT_SCHEMA.iter_errors(doc):
        has_doc_errors = True
        displayable_path = ".".join(map(str, error.absolute_path))
        context = f"({displayable_path}) " if displayable_path else ""
        yield _error("document_schema", f"{context}{error.message} ")

    # The jsonschema error message for this (common error) is garbage. Make it clearer.
    measurements = doc.get("measurements")
    if (measurements is not None) and not isinstance(measurements, Sequence):
        yield _error(
            "measurements_list",
            f"Product measurements should be a list/sequence "
            f"(Found a {type(measurements).__name__!r}).",
        )

    # There's no point checking further if the core doc structure is wrong.
    if has_doc_errors:
        return

    if not doc.get("license", "").strip():
        yield _warning(
            "no_license",
            f"Product {doc['name']!r} has no license field",
            hint='Eg. "CC-BY-4.0" (SPDX format), "various" or "proprietary"',
        )

    # Check measurement name clashes etc.
    if measurements is None:
        # Products don't have to have measurements. (eg. provenance-only products)
        ...
    else:
        seen_names_and_aliases = collections.defaultdict(list)
        for measurement in measurements:
            measurement_name = measurement.get("name")
            dtype = measurement.get("dtype")
            nodata = measurement.get("nodata")
            if not numpy_value_fits_dtype(nodata, dtype):
                yield _error(
                    "unsuitable_nodata",
                    f"Measurement {measurement_name!r} nodata {nodata!r} does not fit a {dtype!r}",
                )

            # Were any of the names seen in other measurements?
            these_names = measurement_name, *measurement.get("aliases", ())
            for new_field_name in these_names:
                measurements_with_this_name = seen_names_and_aliases[new_field_name]
                if measurements_with_this_name:
                    seen_in = " and ".join(
                        repr(s)
                        for s in ([measurement_name, *measurements_with_this_name])
                    )

                    # If the same name is used by different measurements, its a hard error.
                    yield _error(
                        "duplicate_measurement_name",
                        f"Name {new_field_name!r} is used by multiple measurements",
                        hint=f"It's duplicated in an alias. "
                        f"Seen in measurement(s) {seen_in}",
                    )

            # Are any names duplicated within the one measurement? (not an error, but info)
            for duplicate_name in _find_duplicates(these_names):
                yield _info(
                    "duplicate_alias_name",
                    f"Measurement {measurement_name!r} has a duplicate alias named {duplicate_name!r}",
                )

            for field_ in these_names:
                seen_names_and_aliases[field_].append(measurement_name)


def validate_metadata_type(doc: Dict) -> ValidationMessages:
    """
    Check for common metadata-type mistakes
    """

    # Validate it against ODC's schema (there will be refused by ODC otherwise)
    for error in serialise.METADATA_TYPE_SCHEMA.iter_errors(doc):
        displayable_path = ".".join(map(str, error.absolute_path))
        context = f"({displayable_path}) " if displayable_path else ""
        yield _error("document_schema", f"{context}{error.message} ")


def _find_duplicates(values: Iterable[str]) -> Generator[str, None, None]:
    """Return any duplicate values in the given sequence

    >>> list(_find_duplicates(('a', 'b', 'c')))
    []
    >>> list(_find_duplicates(('a', 'b', 'b')))
    ['b']
    >>> list(_find_duplicates(('a', 'b', 'b', 'a')))
    ['a', 'b']
    """
    previous = None
    for v in sorted(values):
        if v == previous:
            yield v
        previous = v


def numpy_value_fits_dtype(value, dtype):
    """
    Can the value be exactly represented by the given numpy dtype?

    >>> numpy_value_fits_dtype(3, 'uint8')
    True
    >>> numpy_value_fits_dtype(3, np.dtype('uint8'))
    True
    >>> numpy_value_fits_dtype(-3, 'uint8')
    False
    >>> numpy_value_fits_dtype(3.5, 'float32')
    True
    >>> numpy_value_fits_dtype(3.5, 'int16')
    False
    >>> numpy_value_fits_dtype(float('NaN'), 'float32')
    True
    >>> numpy_value_fits_dtype(float('NaN'), 'int32')
    False
    """
    dtype = np.dtype(dtype)

    if value is None:
        value = 0

    if _is_nan(value):
        return np.issubdtype(dtype, np.floating)
    else:
        return np.all(np.array([value], dtype=dtype) == [value])


@define
class ExpectedMeasurement:
    name: str
    dtype: str
    nodata: int

    @classmethod
    def from_definition(cls, doc: Dict):
        return ExpectedMeasurement(doc["name"], doc.get("dtype"), doc.get("nodata"))


# Name of a field and its possible offsets in the document.
FieldNameOffsetS = Tuple[str, Set[List[str]]]


def validate_paths(
    paths: List[str],
    thorough: bool = False,
    product_definitions: Dict[str, Dict] = None,
    metadata_type_definitions: Dict[str, Dict] = None,
    expect: ValidationExpectations = None,
) -> Generator[Tuple[str, List[ValidationMessage]], None, None]:
    """Validate the list of paths. Product documents can be specified before their datasets."""

    products = dict(product_definitions or {})
    metadata_types = dict(metadata_type_definitions or {})

    for url, doc, was_specified_by_user in read_paths(paths):
        messages = []
        kind = filename_doc_kind(url)
        if kind is None:
            kind = guess_kind_from_contents(doc)
            if kind and (kind in DOC_TYPE_SUFFIXES):
                # It looks like an ODC doc but doesn't have the standard suffix.
                messages.append(
                    _warning(
                        "missing_suffix",
                        f"Document looks like a {kind.name} but does not have "
                        f'filename extension "{DOC_TYPE_SUFFIXES[kind]}{_readable_doc_extension(url)}"',
                    )
                )

        if kind == DocKind.product:
            messages.extend(validate_product(doc))
            if "name" in doc:
                products[doc["name"]] = doc
        elif kind == DocKind.dataset:
            messages.extend(
                validate_eo3_doc(
                    doc,
                    url,
                    products,
                    metadata_types,
                    thorough,
                    expect=expect,
                )
            )
        elif kind == DocKind.metadata_type:
            messages.extend(validate_metadata_type(doc))
            if "name" in doc:
                metadata_types[doc["name"]] = doc

        # Otherwise it's a file we don't support.
        # If the user gave us the path explicitly, it seems to be an error.
        # (if they didn't -- it was found via scanning directories -- we don't care.)
        elif was_specified_by_user:
            if kind is None:
                raise ValueError(f"Unknown document type for {url}")
            else:
                raise NotImplementedError(
                    f"Cannot currently validate {kind.name} files"
                )
        else:
            # Not a doc type we recognise, and the user didn't specify it. Skip it.
            continue

        yield url, messages


def _get_field_offsets(metadata_type: Dict) -> Iterable[FieldNameOffsetS]:
    """
    Yield all fields and their possible document-offsets that are expected for this metadata type.

    Eg, if the metadata type has a region_code field expected properties->region_code, this
    will yield ('region_code', {['properties', 'region_code']})

    (Properties can have multiple offsets, where ODC will choose the first non-null one, hence the
    return of multiple offsets for each field.)
    """
    dataset_section = metadata_type["dataset"]
    search_fields = dataset_section["search_fields"]

    # The fixed fields of ODC. 'id', 'label', etc.
    for field_ in dataset_section:
        if field_ == "search_fields":
            continue

        offset = dataset_section[field_]
        if offset is not None:
            yield field_, [offset]

    # The configurable search fields.
    for field_, spec in search_fields.items():
        offsets = []
        if "offset" in spec:
            offsets.append(spec["offset"])
        offsets.extend(spec.get("min_offset", []))
        offsets.extend(spec.get("max_offset", []))

        yield field_, offsets


def _readable_doc_extension(uri: str):
    """
    >>> _readable_doc_extension('something.json.gz')
    '.json.gz'
    >>> _readable_doc_extension('something.yaml')
    '.yaml'
    >>> _readable_doc_extension('apple.odc-metadata.yaml.gz')
    '.yaml.gz'
    >>> _readable_doc_extension('products/tmad/tmad_product.yaml#part=1')
    '.yaml'
    >>> _readable_doc_extension('/tmp/human.06.tall.yml')
    '.yml'
    >>> # Not a doc, even though it's compressed.
    >>> _readable_doc_extension('db_dump.gz')
    >>> _readable_doc_extension('/tmp/nothing')
    """
    path = urlparse(uri).path
    compression_formats = (".gz",)
    doc_formats = (
        ".yaml",
        ".yml",
        ".json",
    )
    suffix = "".join(
        s.lower()
        for s in Path(path).suffixes
        if s.lower() in doc_formats + compression_formats
    )
    # If it's only compression, no doc format, it's not valid.
    if suffix in compression_formats:
        return None
    return suffix or None


def read_paths(
    input_paths: Iterable[str],
) -> Generator[Tuple[str, Union[Dict, str], bool], None, None]:
    """
    Read the given input paths, returning a URL, document, and whether
    it was explicitly given by the user.

    When a local directory is specified, inner readable docs are returned, but will
    be marked as not explicitly specified.
    """
    for input_ in input_paths:
        for uri, was_specified in expand_paths_as_uris([input_]):
            try:
                for full_uri, doc in read_documents(uri, uri=True):
                    yield full_uri, doc, was_specified
            except InvalidDocException as e:
                if was_specified:
                    raise
                else:
                    echo(e, err=True)


def expand_paths_as_uris(
    input_paths: Iterable[str],
) -> Generator[Tuple[Path, bool], None, None]:
    """
    For any paths that are directories, find inner documents that are known.

    Returns Tuples: path as a URL, and whether it was specified explicitly by user.
    """
    for input_ in input_paths:
        if is_url(input_):
            yield input_, True
        else:
            path = Path(input_).resolve()
            if path.is_dir():
                for found_path in path.rglob("*"):
                    if _readable_doc_extension(found_path.as_uri()) is not None:
                        yield found_path.as_uri(), False
            else:
                yield path.as_uri(), True


def validate_eo3_doc(
    doc: Dict,
    location: Union[str, Path],
    products: Dict[str, Dict],
    metadata_types: Dict[str, Dict],
    thorough: bool = False,
    expect: ValidationExpectations = None,
) -> List[ValidationMessage]:
    messages = []

    # TODO: follow ODC's match rules?

    matched_product = None

    if products:
        matched_product, messages = _match_product(doc, products)
    else:
        messages.append(
            ValidationMessage(
                Level.error if thorough else Level.info,
                "no_product",
                "No product provided: validating dataset information alone",
            )
        )

    metadata_type = None
    if metadata_types and matched_product:
        metadata_type = matched_product["metadata_type"]
        if metadata_type not in metadata_types:
            messages.append(
                ValidationMessage(
                    Level.error if thorough else Level.info,
                    "no_metadata_type",
                    f"Metadata type not provided {metadata_type}: not validating fields",
                )
            )

    messages.extend(
        validate_dataset(
            doc,
            product_definition=matched_product,
            readable_location=location,
            thorough=thorough,
            metadata_type_definition=metadata_types.get(metadata_type),
            expect=expect,
        )
    )
    return messages


def _get_printable_differences(dict1: Dict, dict2: Dict):
    """
    Get a series of lines to print that show the reason that dict1 is not a superset of dict2
    """
    dict1 = dict(utils.flatten_dict(dict1))
    dict2 = dict(utils.flatten_dict(dict2))

    for path in dict2.keys():
        v1, v2 = dict1.get(path), dict2.get(path)
        if v1 != v2:
            yield f"{path}: {v1!r} != {v2!r}"


def _get_product_mismatch_reasons(dataset_doc: Dict, product_definition: Dict):
    """
    Which fields don't match the given dataset doc to a product definition?

    Gives human-readable lines of text.
    """
    yield from _get_printable_differences(dataset_doc, product_definition["metadata"])


def _match_product(
    dataset_doc: Dict, product_definitions: Dict[str, Dict]
) -> Tuple[Optional[Dict], List[ValidationMessage]]:
    """Match the given dataset to a product definition"""

    product = None

    # EO3 datasets often put the product name directly inside.
    specified_product_name = get_path(dataset_doc, ("product", "name"), default=None)
    specified_product_name = specified_product_name or get_path(
        dataset_doc, ("properties", "odc:product"), default=None
    )

    if specified_product_name and (specified_product_name in product_definitions):
        product = product_definitions[specified_product_name]

    matching_products = {
        name: definition
        for name, definition in product_definitions.items()
        if changes.contains(dataset_doc, definition["metadata"])
    }

    # We we have nothing, give up!
    if (not matching_products) and (not product):
        # Find the product that most closely matches it, to helpfully show the differences!
        closest_product_name = None
        closest_differences = None
        for name, definition in product_definitions.items():
            diffs = tuple(_get_product_mismatch_reasons(dataset_doc, definition))
            if (closest_differences is None) or len(diffs) < len(closest_differences):
                closest_product_name = name
                closest_differences = diffs

        difference_hint = _differences_as_hint(closest_differences)
        return None, [
            _error(
                "unknown_product",
                "Dataset does not match the given products",
                hint=f"Closest match is {closest_product_name}, with differences:"
                f"\n{difference_hint}",
            )
        ]

    messages = []

    if specified_product_name not in matching_products:
        if product:
            difference_hint = _differences_as_hint(
                _get_product_mismatch_reasons(dataset_doc, product)
            )
            messages.append(
                _info(
                    "strange_product_claim",
                    f"Dataset claims to be product {specified_product_name!r}, but doesn't match its fields",
                    hint=f"{difference_hint}",
                )
            )
        else:
            messages.append(
                _info(
                    "unknown_product_claim",
                    f"Dataset claims to be product {specified_product_name!r}, but it wasn't supplied.",
                )
            )

    if len(matching_products) > 1:
        matching_names = ", ".join(matching_products.keys())
        messages.append(
            _error(
                "product_match_clash",
                "Multiple products match the given dataset",
                hint=(
                    f"Maybe you need more fields in the 'metadata' section?\n"
                    f"Claims to be a {specified_product_name!r}, and matches {matching_names!r}"
                    if specified_product_name
                    else f"Maybe you need more fields in the 'metadata' section?\n"
                    f"Matches {matching_names!r}"
                ),
            )
        )
        # (We wont pick one from the bunch here. Maybe they already matched one above to use in continuing validation.)

    # Just like ODC, match rules will rule all. Even if their metadata has a "product_name" field.
    if len(matching_products) == 1:
        [product] = matching_products.values()

    return product, messages


def _differences_as_hint(product_diffs):
    return indent("\n".join(product_diffs), prefix="\t")


def _validate_stac_properties(dataset: DatasetDoc):
    for name, value in dataset.properties.items():
        if name not in dataset.properties.KNOWN_PROPERTIES:
            yield _warning("unknown_property", f"Unknown stac property {name!r}")

        else:
            normaliser = dataset.properties.KNOWN_PROPERTIES.get(name)
            if normaliser and value is not None:
                try:
                    normalised_value = normaliser(value)
                    # A normaliser can return two values, the latter adding extra extracted fields.
                    if isinstance(normalised_value, tuple):
                        normalised_value = normalised_value[0]

                    # It's okay for datetimes to be strings
                    # .. since ODC's own loader does that.
                    if isinstance(normalised_value, datetime) and isinstance(
                        value, str
                    ):
                        value = ciso8601.parse_datetime(value)

                    # Special case for dates, as "no timezone" and "utc timezone" are treated identical.
                    if isinstance(value, datetime):
                        value = default_utc(value)

                    if not isinstance(value, type(normalised_value)):
                        yield _warning(
                            "property_type",
                            f"Value {value} expected to be "
                            f"{type(normalised_value).__name__!r} (got {type(value).__name__!r})",
                        )
                    elif normalised_value != value:
                        if _is_nan(normalised_value) and _is_nan(value):
                            # Both are NaNs, ignore.
                            pass
                        else:
                            yield _warning(
                                "property_formatting",
                                f"Property {value!r} expected to be {normalised_value!r}",
                            )
                except ValueError as e:
                    yield _error("invalid_property", f"{name!r}: {e.args[0]}")

    if "odc:producer" in dataset.properties:
        producer = dataset.properties["odc:producer"]
        # We use domain name to avoid arguing about naming conventions ('ga' vs 'geoscience-australia' vs ...)
        if "." not in producer:
            yield _warning(
                "producer_domain",
                "Property 'odc:producer' should be the organisation's domain name. Eg. 'ga.gov.au'",
            )

    # This field is a little odd, but is expected by the current version of ODC.
    # (from discussion with Kirill)
    if not dataset.properties.get("odc:file_format"):
        yield _warning(
            "global_file_format",
            "Property 'odc:file_format' is empty",
            hint="Usually 'GeoTIFF'",
        )


def _is_nan(v):
    # Due to JSON serialisation, nan can also be represented as a string 'NaN'
    if isinstance(v, str):
        return v == "NaN"
    return isinstance(v, float) and math.isnan(v)


def _validate_geo(dataset: DatasetDoc, expect_geometry: bool = True):
    has_some_geo = _has_some_geo(dataset)
    if not has_some_geo and expect_geometry:
        yield _info("non_geo", "No geo information in dataset")
        return

    if dataset.geometry is None:
        if expect_geometry:
            yield _info("incomplete_geo", "Dataset has some geo fields but no geometry")
    elif not dataset.geometry.is_valid:
        yield _error(
            "invalid_geometry",
            f"Geometry is not a valid shape: {explain_validity(dataset.geometry)!r}",
        )

    # TODO: maybe we'll allow no grids: backwards compat with old metadata.
    if not dataset.grids:
        yield _error("incomplete_grids", "Dataset has some geo fields but no grids")

    if not dataset.crs:
        yield _error("incomplete_crs", "Dataset has some geo fields but no crs")
    else:
        # We only officially support epsg code (recommended) or wkt.
        if dataset.crs.lower().startswith("epsg:"):
            try:
                CRS.from_string(dataset.crs)
            except CRSError as e:
                yield _error("invalid_crs_epsg", e.args[0])

            if dataset.crs.lower() != dataset.crs:
                yield _warning("mixed_crs_case", "Recommend lowercase 'epsg:' prefix")
        else:
            wkt_crs = None
            try:
                wkt_crs = CRS.from_wkt(dataset.crs)
            except CRSError as e:
                yield _error(
                    "invalid_crs",
                    f"Expect either an epsg code or a WKT string: {e.args[0]}",
                )

            if wkt_crs and wkt_crs.is_epsg_code:
                yield _warning(
                    "non_epsg",
                    f"Prefer an EPSG code to a WKT when possible. (Can change CRS to 'epsg:{wkt_crs.to_epsg()}')",
                )


def _has_some_geo(dataset: DatasetDoc) -> bool:
    return dataset.geometry is not None or dataset.grids or dataset.crs


def display_result_console(
    url: str, is_valid: bool, messages: List[ValidationMessage], quiet=False
):
    """
    Print validation messages to the Console (using colour if available).
    """
    # Otherwise console output, with color if possible.
    if messages or not quiet:
        echo(f"{bool_style(is_valid)} {url}")

    for message in messages:
        hint = ""
        if message.hint:
            # Indent the hint if it's multi-line.
            if "\n" in message.hint:
                hint = "\t\tHint:\n"
                hint += indent(message.hint, "\t\t" + (" " * 5))
            else:
                hint = f"\t\t(Hint: {message.hint})"
        s = {
            Level.info: dict(),
            Level.warning: dict(fg="yellow"),
            Level.error: dict(fg="red"),
        }
        displayable_code = style(f"{message.code}", **s[message.level], bold=True)
        echo(f"\t{message.level.name[0].upper()} {displayable_code} {message.reason}")
        if hint:
            echo(hint)


def display_result_github(url: str, is_valid: bool, messages: List[ValidationMessage]):
    """
    Print validation messages using Github Action's command language for warnings/errors.
    """
    echo(f"{bool_style(is_valid)} {url}")
    for message in messages:
        hint = ""
        if message.hint:
            # Indent the hint if it's multi-line.
            if "\n" in message.hint:
                hint = "\n\nHint:\n"
                hint += indent(message.hint, (" " * 5))
            else:
                hint = f"\n\n(Hint: {message.hint})"

        if message.level == Level.error:
            code = "::error"
        else:
            code = "::warning"

        text = f"{message.reason}{hint}"

        # URL-Encode any newlines
        text = text.replace("\n", "%0A")
        # TODO: Get the real line numbers?
        echo(f"{code} file={url},line=1::{text}")


_OUTPUT_WRITERS = dict(
    plain=display_result_console,
    quiet=partial(display_result_console, quiet=True),
    github=display_result_github,
)


@click.command(
    help=__doc__
    + """
Paths can be products, dataset documents, or directories to scan (for files matching
names '*.odc-metadata.yaml' etc), either local or URLs.

Datasets are validated against matching products that have been scanned already, so specify
products first, and datasets later, to ensure they can be matched.
"""
)
@click.version_option()
@click.argument("paths", nargs=-1)
@click.option(
    "--warnings-as-errors",
    "-W",
    "strict_warnings",
    is_flag=True,
    help="Fail if any warnings are produced",
)
@click.option(
    "-f",
    "--output-format",
    help="Output format",
    type=click.Choice(list(_OUTPUT_WRITERS)),
    # Are we in Github Actions?
    # Send any warnings/errors in its custom format
    default="github" if "GITHUB_ACTIONS" in os.environ else "plain",
    show_default=True,
)
@click.option(
    "--thorough",
    is_flag=True,
    help="Attempt to read the data/measurements, and check their properties match",
)
@click.option(
    "--explorer-url",
    "explorer_url",
    help="Use product definitions from the given Explorer URL to validate datasets. "
    'Eg: "https://explorer.dea.ga.gov.au/"',
)
@click.option(
    "--odc",
    "use_datacube",
    is_flag=True,
    help="Use product definitions from datacube to validate datasets",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    default=False,
    help="Only print problems, one per line",
)
def run(
    paths: List[str],
    strict_warnings,
    quiet,
    thorough: bool,
    explorer_url: str,
    use_datacube: bool,
    output_format: str,
):
    expect = ValidationExpectations()
    validation_counts: Counter[Level] = collections.Counter()
    invalid_paths = 0
    current_location = Path(".").resolve().as_uri() + "/"

    product_definitions = _load_remote_product_definitions(use_datacube, explorer_url)

    if output_format == "plain" and quiet:
        output_format = "quiet"
    write_file_report = _OUTPUT_WRITERS[output_format]

    for url, messages in validate_paths(
        paths,
        thorough=thorough,
        expect=expect,
        product_definitions=product_definitions,
    ):
        if url.startswith(current_location):
            url = url[len(current_location) :]

        levels = collections.Counter(m.level for m in messages)
        is_invalid = levels[Level.error] > 0
        if strict_warnings:
            is_invalid |= levels[Level.warning] > 0

        if quiet:
            # Errors/Warnings only. Remove info-level.
            messages = [m for m in messages if m.level != Level.info]

        if is_invalid:
            invalid_paths += 1

        for message in messages:
            validation_counts[message.level] += 1

        write_file_report(
            url=url,
            is_valid=not is_invalid,
            messages=messages,
        )

    # Print a summary on stderr for humans.
    if not quiet:
        result = (
            style("failure", fg="red", bold=True)
            if invalid_paths > 0
            else style("valid", fg="green", bold=True)
        )
        secho(f"\n{result}: ", nl=False, err=True)
        if validation_counts:
            echo(
                ", ".join(
                    f"{v} {k.name}{'s' if v > 1 else ''}"
                    for k, v in validation_counts.items()
                ),
                err=True,
            )
        else:
            secho(f"{len(paths)} paths", err=True)

    sys.exit(invalid_paths)


def _load_remote_product_definitions(
    from_datacube: bool = False,
    from_explorer_url: Optional[str] = None,
) -> Dict[str, Dict]:
    product_definitions = {}
    # Load any remote products that were asked for.
    if from_explorer_url:
        for definition in _load_explorer_product_definitions(from_explorer_url):
            product_definitions[definition["name"]] = definition
        secho(f"{len(product_definitions)} Explorer products", err=True)

    if from_datacube:
        # The normal datacube environment variables can be used to choose alternative configs.
        with Datacube(app="eo3-validate") as dc:
            for product in dc.index.products.get_all():
                product_definitions[product.name] = product.definition

        secho(f"{len(product_definitions)} ODC products", err=True)
    return product_definitions


def _load_doc(url):
    return list(load_documents(url))


def _load_explorer_product_definitions(
    explorer_url: str,
    workers: int = 6,
) -> Generator[Dict, None, None]:
    """
    Read all product yamls from the given Explorer instance,

    eg: https://explorer.dea.ga.gov.au/products/ls5_fc_albers.odc-product.yaml
    """
    product_urls = [
        urljoin(explorer_url, f"/products/{name.strip()}.odc-product.yaml")
        for name in urlopen(urljoin(explorer_url, "products.txt"))  # nosec
        .read()
        .decode("utf-8")
        .split("\n")
    ]
    count = 0
    with multiprocessing.Pool(workers) as pool:
        for product_definitions in pool.imap_unordered(_load_doc, product_urls):
            count += 1
            echo(f"\r{count} Explorer products", nl=False)
            yield from product_definitions
        pool.close()
        pool.join()
    echo()



import click

from ..prepare.landsat_l1_prepare import main as landsat_l1
from ..prepare.nasa_c_m_mcd43a1_6_prepare import main as mcd43a1
from ..prepare.noaa_c_c_prwtreatm_1_prepare import main as prwtr
from ..prepare.sentinel_l1_prepare import main as sentinel_l1


@click.group()
@click.version_option()
def run():
    pass


run.add_command(landsat_l1, name="landsat-l1")
run.add_command(sentinel_l1, name="sentinel-l1")
run.add_command(mcd43a1, name="modis-mcd43a1")
run.add_command(prwtr, name="noaa-prwtr")


run.add_command(sentinel_l1, name="sentinel-l1c")

if __name__ == "__main__":
    run()



import re
import sqlite3
from pathlib import Path
from pprint import pprint
from typing import Dict, Iterable, List, Optional, Tuple

import click
from attr import define
from click import echo, pass_obj

from eodatasets3.ui import PathPath

NCI_L1C_LOCATION = Path("/g/data/fj7/Copernicus/Sentinel-2/MSI/L1C")

DEFAULT_DB = Path(__file__).parent / "s2_regions.db"


_AREA_PARTS = re.compile(r"(\d+)([NS])(\d+)([EW])")


def area_to_tuple(area: str) -> Tuple[int, int, int, int]:
    """
    >>> area_to_tuple('20S120E-25S125E')
    (-20, 120, -25, 125)
    >>> area_to_tuple('00N160W-05S155W')
    (0, -160, -5, -155)
    """

    def point(part: str) -> Tuple[int, int]:
        match = _AREA_PARTS.match(part)
        if match is None:
            raise ValueError(f"Not an area? {part!r} to {_AREA_PARTS!r}")
        lat, ns, lon, ew = match.groups()
        if ns == "S":
            lat = "-" + lat
        if ew == "W":
            lon = "-" + lon

        return int(lat), int(lon)

    first, second = area.split("-")
    return (*point(first), *point(second))


@define
class FolderInfo:
    """
    Information extracted from a standard S2 folder layout.
    """

    year: int
    month: int
    area: str
    region_code: Optional[str]

    # Compiled regexp for extracting year, month and region
    # Standard layout is of the form: 'L1C/{yyyy}/{yyyy}-{mm}/{area}/S2*_{region}_{timestamp}(.zip)'
    STANDARD_SUBFOLDER_LAYOUT = re.compile(
        r"(\d{4})/(\d{4})-(\d{2})/([\dNESW]+-[\dNESW]+)/"
        r"S2[AB](?:_OPER_PRD)?_MSIL1C(?:_PDMC)?(?:_[a-zA-Z0-9]+){3}(?:_T([A-Z\d]+))?_[\dT]+(\.zip|/tileInfo\.json)?$"
    )

    @property
    def area_tuple(self):
        return area_to_tuple(self.area)

    @classmethod
    def for_path(cls, path: Path) -> Optional["FolderInfo"]:
        """
        Can we extract information from the given path?

        Returns None if we can't.
        """
        m = cls.STANDARD_SUBFOLDER_LAYOUT.search(path.as_posix())
        if not m:
            return None

        year, year2, month, area, region_code, extension = m.groups()
        if year != year2:
            raise ValueError(f"Year mismatch in {path}")

        return FolderInfo(int(year), int(month, 10), area, region_code)


class RegionLookup:
    """
    A lookup table for S2 regions and areas.

    This is for matching areas to region_codes on older S2 data that does not contain
    a region_code.

    The table is stored as an sqlite database.
    """

    def __init__(self, db_path: Path = None):
        self.db_path = db_path
        self._db: Optional[sqlite3.Connection] = None

    def open(self):
        if self._db is None:
            if self.db_path is None:
                # Open the default db read-only
                self._db = sqlite3.connect(f"{DEFAULT_DB.as_uri()}?mode=ro", uri=True)
            else:
                self._db = sqlite3.connect(self.db_path)

    def close(self):
        if self._db:
            self._db.close()
            self._db = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add_from_file(self, paths_file: Path) -> int:
        """Add a paths to the lookup table"""
        opener = open
        if paths_file.suffix.lower() == ".gz":
            import gzip

            opener = gzip.open

        return self.add_from_paths(Path(p.strip()) for p in opener(paths_file, "rt"))

    def add_from_scanning_path(self, l1cs: Path) -> int:
        """Scan for zips in a directory and add them to the table"""
        return self.add_from_paths(l1cs.rglob("*.zip"))

    def add_from_paths(self, paths: Iterable[Path]):
        """Add the given pathsto the lookup table"""
        self.open()
        s = self._db.cursor()
        s.execute(
            """
            create table if not exists regions (
                lat1 integer not null,
                lon1 integer not null,
                lat2 integer not null,
                lon2 integer not null,
                region_code text null,
                primary key (lat1, lon1, lat2, lon2, region_code)
            )
        """
        )
        s.execute("begin")

        def vals(path: Path):
            i = FolderInfo.for_path(path)
            return (
                *i.area_tuple,
                i.region_code,
            )

        res = s.executemany(
            "insert into regions values (?,?,?,?,?) on conflict do nothing",
            (vals(path) for path in paths),
        )
        insert_count = res.rowcount
        s.execute("commit")
        return insert_count

    def get(self, area: str) -> List[str]:
        """Get known region codes for an area"""
        self.open()

        def get_lazy():
            for [code] in self._db.execute(
                """
                    select distinct region_code from regions where (lat1, lon1, lat2, lon2)  = (?, ?, ?, ?)
                """,
                area_to_tuple(area),
            ):
                yield code

        return list(get_lazy())

    def vacuum(self):
        """Vacuum the database (reclaim space)"""
        self.open()
        self._db.execute("vacuum")

    def stats(self) -> Dict:
        """Get stats about the lookup table"""
        self.open()
        res = self._db.execute(
            """
            select
                count(*) as total,
                count(distinct region_code) as unique_regions
                --- count(distinct (lat1, lon1, lat2, lon2)) as unique_areas
            from regions
        """
        ).fetchone()

        return res


@click.group("s2_regions", help=__doc__)
@click.option("--db", default=DEFAULT_DB, type=PathPath())
@click.pass_context
def cli(ctx, db: Path):
    ctx.obj = RegionLookup(db)


@cli.command("create", help="Recreate the database")
@click.option("--scan-path", default=None, type=PathPath(exists=True))
@click.option("-f", "paths_file", default=None, type=PathPath(exists=True))
@pass_obj
def cli_create(db: RegionLookup, scan_path: Path, paths_file: Path):
    if scan_path is None and paths_file is None:
        echo("Nothing specified. Scanning default NCI location.")
        scan_path = NCI_L1C_LOCATION

    if scan_path is not None:
        echo("Scanning....")
        inserted = db.add_from_scanning_path(scan_path)
        echo(f"Found {inserted} items in {scan_path}")

    if paths_file is not None:
        echo("Adding from file....")
        inserted = db.add_from_file(paths_file)
        echo(f"Loaded {inserted} items from file {paths_file}")

    db.vacuum()
    db.close()


@cli.command("get", help="Get region codes for an area")
@click.argument("area", type=str)
@click.pass_obj
def cli_get(db: RegionLookup, area: str):
    for region in db.get(area):
        echo(region)

    db.close()


@cli.command("info", help="Get region codes for an area")
@click.pass_obj
def cli_info(
    db: RegionLookup,
):
    pprint(db.stats())
    db.close()


if __name__ == "__main__":
    cli()



import uuid
from pathlib import Path

import rasterio
from shapely.geometry import box

from eodatasets3 import DatasetPrepare, GridSpec

_ESRI_ID_NAMESPACE = uuid.UUID("88a620e0-6b62-462b-9b23-4027e05898f3")


def as_eo3(uri: str):
    """
    Prepare an EO3 doc for the given (possibly remote ULR) ESRI TIF image.

    This is an EOD3-equivalent of ``odc-tools/apps/dc_tools/odc/apps/dc_tools/esri_land_cover_to_dc.py``

    Added mainly for testing a different use-case of the EOD3 APIs.
    """
    path = Path(uri)

    with rasterio.open(uri, GEOREF_SOURCES="INTERNAL") as opened_asset:
        grid_spec = GridSpec.from_rio(opened_asset)
        bbox = opened_asset.bounds
        nodata = opened_asset.nodata

    region, date_range = path.stem.split("_")
    start_date, end_date = date_range.split("-")
    if grid_spec.crs is None:
        print(f"Empty CRS: {uri}")
        return None

    # geom = (
    #     Geometry(box(*bbox), crs=grid_spec.crs)
    #     .to_crs("EPSG:4326", wrapdateline=True)
    #     .to_crs(grid_spec.crs)
    # )

    d = DatasetPrepare(dataset_location=uri)
    d.dataset_id = uuid.uuid5(_ESRI_ID_NAMESPACE, path.name)
    d.geometry = box(*bbox)

    d.product_name = "esri_land_cover"
    d.datetime = start_date
    d.datetime_range = start_date, end_date
    d.region_code = region.strip()
    d.processed_now()

    d.note_measurement(
        "classification",
        uri,
        relative_to_dataset_location=True,
        grid=grid_spec,
        expand_valid_data=False,
        nodata=nodata,
    )

    return d.to_dataset_doc(embed_location=True, validate_correctness=False)



"""
Prepare eo3 metadata for USGS Landsat Level 1 data.

Input dataset paths can be directories or tar files.
"""

import logging
import os
import re
import tarfile
import uuid
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union

import click
import rasterio

from eodatasets3 import DatasetPrepare, serialise, utils
from eodatasets3.properties import FileFormat
from eodatasets3.ui import PathPath

_COPYABLE_MTL_FIELDS = {}
_COPYABLE_MTL_FIELDS["C1"] = [
    (
        "metadata_file_info",
        (
            "landsat_scene_id",
            "landsat_product_id",
            "station_id",
            "processing_software_version",
        ),
    ),
    (
        "product_metadata",
        ("data_type", "ephemeris_type", "wrs_path", "wrs_row", "collection_category"),
    ),
    (
        "image_attributes",
        (
            "ground_control_points_version",
            "ground_control_points_model",
            "geometric_rmse_model_x",
            "geometric_rmse_model_y",
            "ground_control_points_verify",
            "geometric_rmse_verify",
        ),
    ),
    ("projection_parameters", ("scan_gap_interpolation",)),
]
_COPYABLE_MTL_FIELDS["C2"] = [
    (
        "level2_processing_record",
        (
            "landsat_scene_id",
            "landsat_product_id",
            "processing_software_version",
            "algorithm_source_surface_reflectance",
            "collection_category",
            "ground_control_points_version",
            "ground_control_points_model",
            "geometric_rmse_model_x",
            "geometric_rmse_model_y",
        ),
    ),
    (
        "level1_processing_record",
        (
            "landsat_scene_id",
            "landsat_product_id",
            "processing_software_version",
            "collection_category",
            "ground_control_points_version",
            "ground_control_points_model",
            "geometric_rmse_model_x",
            "geometric_rmse_model_y",
        ),
    ),
    (
        "image_attributes",
        (
            "station_id",
            "wrs_path",
            "wrs_row",
            # "ground_control_points_verify",  # not in the test data for C1 or C2
            # "geometric_rmse_verify",   # not in the test data for C1 or C2
        ),
    ),
    # not in the test data for C1 or C2
    # ("level1_projection_parameters", ("scan_gap_interpolation",)),
]

# Static namespace to generate uuids for datacube indexing
USGS_UUID_NAMESPACE = uuid.UUID("276af61d-99f8-4aa3-b2fb-d7df68c5e28f")

LANDSAT_OLI_TIRS_BAND_ALIASES = {
    "band_1": "coastal_aerosol",
    "band_2": "blue",
    "band_3": "green",
    "band_4": "red",
    "band_5": "nir",
    "band_6": "swir_1",
    "band_7": "swir_2",
    "band_8": "panchromatic",
    "band_9": "cirrus",
    "band_st_b10": "lwir",  # USGS only
    "band_10": "lwir_1",
    "band_11": "lwir_2",
    "quality_l1_pixel": "quality",
    # Older collection called quality a "band"
    "band_quality": "quality",
    "qa_aerosol": "qa_aerosol",
    "thermal_radiance": "thermal_radiance",
    "upwell_radiance": "upwell_radiance",
    "downwell_radiance": "downwell_radiance",
    "atmospheric_transmittance": "atmos_transmittance",
    "atmospheric_opacity": "atmos_opacity",
    "emissivity": "emissivity",
    "emissivity_stdev": "emissivity_stdev",
    "cloud_distance": "cloud_distance",
    "quality_l2_aerosol": "qa_aerosol",
    "quality_l2_surface_reflectance_cloud": "qa_cloud",
    "quality_l2_surface_temperature": "qa_temperature",
    "quality_l1_radiometric_saturation": "qa_radsat",
    "angle_sensor_azimuth_band_4": "view_azimuth",
    "angle_sensor_zenith_band_4": "view_zenith",
    "angle_solar_azimuth_band_4": "solar_azimuth",
    "angle_solar_zenith_band_4": "solar_zenith",
}

LANDSAT_xTM_BAND_ALIASES = {
    "band_1": "blue",
    "band_2": "green",
    "band_3": "red",
    "band_4": "nir",
    "band_5": "swir_1",
    "band_6": "tir",
    "band_6_vcid_1": "tir_1",
    "band_6_vcid_2": "tir_2",
    "band_st_b6": "lwir",  # USGS only
    "band_7": "swir_2",
    "band_8": "panchromatic",
    "band_quality": "quality",
    "quality_l1_pixel": "quality",
    "cloud_qa": "qa_cloud",
    "quality_l1_radiometric_saturation": "qa_radsat",
    "quality_l2_aerosol": "qa_aerosol",
    "quality_l2_surface_reflectance_cloud": "qa_cloud",
    "quality_l2_surface_temperature": "qa_temperature",
    "thermal_radiance": "thermal_radiance",
    "upwell_radiance": "upwell_radiance",
    "downwell_radiance": "downwell_radiance",
    "atmospheric_transmittance": "atmos_transmittance",
    "atmospheric_opacity": "atmos_opacity",
    "emissivity": "emissivity",
    "emissivity_stdev": "emissivity_stdev",
    "cloud_distance": "cloud_distance",
    "angle_sensor_azimuth_band_4": "view_azimuth",
    "angle_sensor_zenith_band_4": "view_zenith",
    "angle_solar_azimuth_band_4": "solar_azimuth",
    "angle_solar_zenith_band_4": "solar_zenith",
}

MTL_PAIRS_RE = re.compile(r"(\w+)\s=\s(.*)")

LANDSAT_MTL_MAP = {
    "C1": {
        "product_contents_cn": "metadata_file_info",
        "product_contents_of": "product_metadata",
        "product_contents_fn": "product_metadata",
        "image_attributes": "product_metadata",
        "level1_processing_record": "metadata_file_info",
        "level1_projection_parameters": "projection_parameters",
    },
    "C2": {
        "product_contents_cn": "product_contents",
        "product_contents_of": [
            "level2_processing_record",
            "level1_processing_record",
        ],
        "product_contents_fn": "product_contents",
        "image_attributes": "image_attributes",
        "leveln_processing_record": [
            "level2_processing_record",
            "level1_processing_record",
        ],
        "leveln_projection_parameters": [
            "projection_attributes",
            "level2_projection_parameters",
            "level1_projection_parameters",
        ],
    },
}


def get_band_alias_mappings(sat: str, instrument: str) -> Dict[str, str]:
    """
    To load the band_names for referencing either LANDSAT8 or LANDSAT7 or LANDSAT5 bands
    Landsat7 and Landsat5 have same band names

    >>> get_band_alias_mappings('landsat-8', 'OLI_TIRS') == LANDSAT_OLI_TIRS_BAND_ALIASES
    True
    >>> get_band_alias_mappings('landsat-8', 'OLI') == LANDSAT_OLI_TIRS_BAND_ALIASES
    True
    >>> get_band_alias_mappings('landsat-5', 'TM') == LANDSAT_xTM_BAND_ALIASES
    True
    >>> get_band_alias_mappings('landsat-5', 'TM') == LANDSAT_xTM_BAND_ALIASES
    True
    >>> get_band_alias_mappings('aqua', 'MODIS') == LANDSAT_xTM_BAND_ALIASES
    Traceback (most recent call last):
    ...
    NotImplementedError: Unexpected satellite. Only landsat handled currently. Got 'aqua'
    >>> get_band_alias_mappings('landsat-5', 'MSS') == LANDSAT_xTM_BAND_ALIASES
    Traceback (most recent call last):
    ...
    NotImplementedError: Landsat version not yet supported: 'landsat-5', 'MSS'
    """

    if not sat.startswith("landsat-"):
        raise NotImplementedError(
            f"Unexpected satellite. Only landsat handled currently. Got {sat!r}"
        )
    landsat_number = int(sat.split("-")[1])

    if landsat_number in (8, 9):
        return LANDSAT_OLI_TIRS_BAND_ALIASES
    if landsat_number in (4, 5, 7) and instrument.endswith("TM"):
        return LANDSAT_xTM_BAND_ALIASES

    raise NotImplementedError(
        f"Landsat version not yet supported: {sat!r}, {instrument!r}"
    )


def get_mtl_content(acquisition_path: Path, root_element=None) -> Tuple[Dict, str, str]:
    """
    Find MTL file for the given path. It could be a directory or a tar file.

    It will return the MTL parsed as a dict and its filename.
    """

    def iter_tar_members(tp: tarfile.TarFile) -> Generator[tarfile.TarInfo, None, None]:
        """
        This is a lazy alternative to TarInfo.getmembers() that only reads one tar item at a time.

        We're reading the MTL file, which is almost always the first entry in the tar, and then
        closing it, so we're avoiding skipping through the entirety of the tar.
        """
        member = tp.next()
        while member is not None:
            yield member
            member = tp.next()

    if not acquisition_path.exists():
        raise RuntimeError(f"Missing path '{acquisition_path}'")

    if acquisition_path.is_file() and tarfile.is_tarfile(str(acquisition_path)):
        with tarfile.open(str(acquisition_path), "r") as tp:
            for member in iter_tar_members(tp):
                if "_MTL.txt" in member.name:
                    with tp.extractfile(member) as fp:
                        mtl_doc, file_root_element = read_mtl(fp)
                        return mtl_doc, file_root_element, member.name
            else:
                raise RuntimeError(f"MTL file not found in {acquisition_path!s}")

    else:
        paths = list(acquisition_path.rglob("*_MTL.txt"))
        if not paths:
            raise RuntimeError("No MTL file")
        if len(paths) > 1:
            raise RuntimeError(
                f"Multiple MTL files found in given acq path {acquisition_path}"
            )
        [path] = paths
        with path.open("r") as fp:
            mtl_doc, file_root_element = read_mtl(fp, root_element)
            return mtl_doc, file_root_element, path.name


def read_mtl(fp: Iterable[Union[str, bytes]], root_element=None) -> Tuple[Dict, str]:
    def _parse_value(s: str) -> Union[int, float, str]:
        """
        >>> _parse_value("asdf")
        'asdf'
        >>> _parse_value("123")
        123
        >>> _parse_value("3.14")
        3.14
        """
        s = s.strip('"')
        for parser in [int, float]:
            try:
                return parser(s)
            except ValueError:
                pass
        return s

    def _parse_group(
        lines: Iterable[Union[str, bytes]],
        key_transform: Callable[[str], str] = lambda s: s.lower(),
    ) -> dict:
        tree = {}
        for line in lines:
            # If line is bytes-like convert to str
            if isinstance(line, bytes):
                line = line.decode("utf-8")
            match = MTL_PAIRS_RE.findall(line)
            if match:
                key, value = match[0]
                if key == "GROUP":
                    tree[key_transform(value)] = _parse_group(lines)
                elif key == "END_GROUP":
                    break
                else:
                    tree[key_transform(key)] = _parse_value(value)
        return tree

    tree = _parse_group(fp)
    if root_element is None:
        root_element = next(iter(tree.keys()))
    return tree[root_element], root_element


def _iter_image_paths(product_doc: Dict) -> Generator[Tuple[str, str], None, None]:
    file_pattern = re.compile(r"file_name_([\w]+)")
    for name, filepath in product_doc.items():
        match = file_pattern.match(name)
        if not match or not filepath.endswith(".TIF"):
            continue

        file_id = match.groups()[0]
        yield file_id, filepath


def prepare_and_write(
    ds_path: Path,
    output_yaml_path: Path,
    source_telemetry: Path = None,
    # TODO: Can we infer producer automatically? This is bound to cause mistakes othewise
    producer="usgs.gov",
    embed_location: bool = False,
) -> Tuple[uuid.UUID, Path]:
    """
    Prepare an eo3 metadata file for a Level1 dataset.

    Input dataset path can be a folder or a tar file.
    """
    mtl_doc, root_element, mtl_filename = get_mtl_content(ds_path)
    if not mtl_doc:
        raise ValueError(f"No MTL file found for {ds_path}")
    collection_key = "C2" if root_element == "landsat_metadata_file" else "C1"
    leveln_key_prefix = "leveln" if collection_key == "C2" else "level1"
    coll_map = LANDSAT_MTL_MAP[collection_key]
    usgs_collection_number = mtl_doc[coll_map["product_contents_cn"]].get(
        "collection_number"
    )
    if usgs_collection_number is None:
        raise NotImplementedError(
            "Dataset has no collection number: pre-collection data is not supported."
        )

    data_format = None
    if isinstance(coll_map["product_contents_of"], list):
        for leveln in coll_map["product_contents_of"]:
            if leveln in mtl_doc:
                data_format = mtl_doc[leveln]["output_format"]
                break
    else:
        data_format = mtl_doc[coll_map["product_contents_of"]]["output_format"]
    if data_format.upper() != "GEOTIFF":
        raise NotImplementedError(f"Only GTiff currently supported, got {data_format}")
    file_format = FileFormat.GeoTIFF

    # Assumed below.
    projection_params = None
    if isinstance(coll_map[leveln_key_prefix + "_projection_parameters"], list):
        for leveln in coll_map[leveln_key_prefix + "_projection_parameters"]:
            if leveln in mtl_doc:
                projection_params = mtl_doc[leveln]
                break
    else:
        projection_params = mtl_doc[
            coll_map[leveln_key_prefix + "_projection_parameters"]
        ]
    if (
        "grid_cell_size_thermal" in projection_params
        and "grid_cell_size_reflective" in projection_params
        and (
            projection_params["grid_cell_size_reflective"]
            != projection_params["grid_cell_size_thermal"]
        )
    ):
        raise NotImplementedError("reflective and thermal have different cell sizes")
    ground_sample_distance = min(
        value
        for name, value in projection_params.items()
        if name.startswith("grid_cell_size_")
    )

    leveln_product_id = None
    leveln_processed = None
    leveln_landsat_data_type = None
    if isinstance(coll_map[leveln_key_prefix + "_processing_record"], list):
        for leveln in coll_map[leveln_key_prefix + "_processing_record"]:
            if leveln in mtl_doc:
                leveln_product_id = mtl_doc[leveln]["landsat_product_id"]
                leveln_processed = mtl_doc[leveln]["date_product_generated"]
                leveln_landsat_data_type = mtl_doc[leveln]["processing_level"]
                break
    else:
        leveln_product_id = mtl_doc[coll_map[leveln_key_prefix + "_processing_record"]][
            "landsat_product_id"
        ]
        leveln_processed = mtl_doc[coll_map[leveln_key_prefix + "_processing_record"]][
            "file_date"
        ]  # for C1 only
        leveln_landsat_data_type = mtl_doc[coll_map["product_contents_of"]]["data_type"]

    with DatasetPrepare(
        metadata_path=output_yaml_path,
        dataset_location=ds_path,
        # Detministic ID based on USGS's product id (which changes when the scene is reprocessed by them)
        dataset_id=uuid.uuid5(USGS_UUID_NAMESPACE, leveln_product_id),
        naming_conventions="dea",
    ) as p:
        if source_telemetry:
            if producer != "ga.gov.au":
                raise NotImplementedError(
                    "Only GA's L1 data is expected to have telemetry source data?"
                )
            p.add_source_path(source_telemetry)

        p.platform = mtl_doc[coll_map["image_attributes"]]["spacecraft_id"]
        p.instrument = mtl_doc[coll_map["image_attributes"]]["sensor_id"]
        p.product_family = "level" + leveln_landsat_data_type[1]
        p.producer = producer
        p.datetime = "{}T{}".format(
            mtl_doc[coll_map["image_attributes"]]["date_acquired"],
            mtl_doc[coll_map["image_attributes"]]["scene_center_time"],
        )
        p.processed = leveln_processed
        if collection_key == "C2":
            p.properties["landsat:data_type"] = leveln_landsat_data_type
        p.properties["odc:file_format"] = file_format
        p.properties["eo:gsd"] = ground_sample_distance
        cloud_cover = mtl_doc["image_attributes"]["cloud_cover"]
        # Cloud cover is -1 when missing (such as TIRS-only data)
        if cloud_cover != -1:
            p.properties["eo:cloud_cover"] = cloud_cover
        p.properties["eo:sun_azimuth"] = mtl_doc["image_attributes"]["sun_azimuth"]
        p.properties["eo:sun_elevation"] = mtl_doc["image_attributes"]["sun_elevation"]
        p.properties["landsat:collection_number"] = usgs_collection_number
        for section, fields in _COPYABLE_MTL_FIELDS[collection_key]:
            if section in mtl_doc:
                for field in fields:
                    value = mtl_doc[section].get(field)
                    if (
                        value is not None
                        and p.properties.get(f"landsat:{field}") is None
                    ):
                        p.properties[f"landsat:{field}"] = value

        p.region_code = f"{p.properties['landsat:wrs_path']:03d}{p.properties['landsat:wrs_row']:03d}"
        org_collection_number = utils.get_collection_number(
            p.platform, p.producer, p.properties["landsat:collection_number"]
        )
        p.dataset_version = f"{org_collection_number}.0.{p.processed:%Y%m%d}"

        # NRT product?
        # Category is one of: T1, T2 or RT ('real time')
        if p.properties["landsat:collection_category"] == "RT":
            p.properties["dea:dataset_maturity"] = "nrt"

        band_aliases = get_band_alias_mappings(p.platform, p.instrument)

        for usgs_file_type, file_location in _iter_image_paths(
            mtl_doc[coll_map["product_contents_fn"]]
        ):
            if usgs_file_type not in band_aliases:
                all_found = dict(
                    _iter_image_paths(mtl_doc[coll_map["product_contents_fn"]])
                )
                raise ValueError(
                    f"Band name {usgs_file_type!r} is not known among our aliases. "
                    f"(All bands found in the dataset: {all_found!r})"
                )
            p.note_measurement(
                band_aliases[usgs_file_type],
                file_location,
                relative_to_dataset_location=True,
                expand_valid_data=(
                    usgs_file_type.startswith("band_")
                    and (
                        # The older collection called quality a "band"
                        "quality" not in usgs_file_type
                    )
                ),
            )

        p.note_accessory_file("metadata:landsat_mtl", Path(mtl_filename))
        return p.done(embed_location=embed_location)


@click.command(help=__doc__)
@click.option(
    "--output-base",
    help="Write metadata files into a directory instead of alongside each dataset",
    required=False,
    type=PathPath(exists=True, writable=True, dir_okay=True, file_okay=False),
)
@click.option(
    "--source",
    "source_telemetry",
    help="Path to the source telemetry data for all of the provided datasets"
    "(either the folder or metadata file)",
    required=False,
    type=PathPath(exists=True),
)
@click.option(
    "--embed-location/--no-embed-location",
    is_flag=True,
    help="Embed the location of the dataset in the metadata "
    "(if you wish to store them separately)",
)
@click.option(
    "--producer",
    help="Organisation that produced the data: probably either 'ga.gov.au' or 'usgs.gov'.",
    required=False,
    default="usgs.gov",
)
@click.argument(
    "datasets", type=PathPath(exists=True, readable=True, writable=False), nargs=-1
)
@click.option(
    "--overwrite-existing/--skip-existing",
    is_flag=True,
    default=False,
    help="Overwrite if exists (otherwise skip)",
)
@click.option(
    "--newer-than",
    type=serialise.ClickDatetime(),
    default=None,
    help="Only process files newer than this date",
)
def main(
    output_base: Optional[Path],
    datasets: List[Path],
    overwrite_existing: bool,
    producer: str,
    embed_location: bool,
    source_telemetry: Optional[Path],
    newer_than: datetime,
):
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
    )
    with rasterio.Env():
        for ds in datasets:
            if output_base:
                output = output_base.absolute().joinpath(
                    *utils.subfolderise(_dataset_region_code(ds))
                )
                output.mkdir(parents=True, exist_ok=True)
            else:
                # Alongside the dataset itself.
                output = ds.absolute().parent

            ds_path = _normalise_dataset_path(Path(ds).absolute())
            (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime) = os.stat(ds)
            create_date = datetime.utcfromtimestamp(ctime)
            if newer_than and (create_date <= newer_than):
                logging.info(
                    "Creation time %s older than start date %s ...SKIPPING %s",
                    newer_than - create_date,
                    newer_than,
                    ds_path.name,
                )
                continue

            logging.info("Processing %s", ds_path)
            output_yaml = output / f"{_dataset_name(ds_path)}.odc-metadata.yaml"

            if output_yaml.exists():
                if not overwrite_existing:
                    logging.info("Output exists: skipping. %s", output_yaml)
                    continue

                logging.info("Output exists: overwriting %s", output_yaml)

            output_uuid, output_path = prepare_and_write(
                ds_path,
                output_yaml,
                producer=producer,
                source_telemetry=source_telemetry,
                embed_location=embed_location,
            )
            logging.info("Wrote dataset %s to %s", output_uuid, output_path)


def _normalise_dataset_path(input_path: Path) -> Path:
    """
    Dataset path should be either the direct imagery folder (mtl+bands) or a tar path.

    Translate other inputs (example: the MTL path) to one of the two.

    >>> import tempfile
    >>> tmppath = Path(tempfile.mkdtemp())
    >>> ds_path = tmppath.joinpath('LE07_L1GT_104078_20131209_20161119_01_T1')
    >>> ds_path.mkdir()
    >>> mtl_path = ds_path / 'LC08_L1TP_090084_20160121_20170405_01_T1_MTL.txt'
    >>> mtl_path.write_text('<mtl content>')
    13
    >>> _normalise_dataset_path(ds_path).relative_to(tmppath).as_posix()
    'LE07_L1GT_104078_20131209_20161119_01_T1'
    >>> _normalise_dataset_path(mtl_path).relative_to(tmppath).as_posix()
    'LE07_L1GT_104078_20131209_20161119_01_T1'
    >>> tar_path = tmppath / 'LS_L1GT.tar.gz'
    >>> tar_path.write_text('fake tar')
    8
    >>> _normalise_dataset_path(tar_path).relative_to(tmppath).as_posix()
    'LS_L1GT.tar.gz'
    >>> _normalise_dataset_path(Path(tempfile.mkdtemp()))
    Traceback (most recent call last):
    ...
    ValueError: No MTL files within input path .... Not a dataset?
    """
    input_path = normalise_nci_symlinks(input_path)
    if input_path.is_file():
        if ".tar" in input_path.suffixes:
            return input_path
        input_path = input_path.parent

    mtl_files = list(input_path.rglob("*_MTL.txt"))

    if not mtl_files:
        raise ValueError(
            f"No MTL files within input path '{input_path}'. Not a dataset?"
        )
    if len(mtl_files) > 1:
        raise ValueError(
            f"Multiple MTL files in a single dataset (got path: {input_path})"
        )
    return input_path


def normalise_nci_symlinks(input_path: Path) -> Path:
    """
    If it's an NCI lustre path, always use the symlink (`/g/data`) rather than specific drives (eg. `/g/data2`).

    >>> normalise_nci_symlinks(Path('/g/data2/v10/some/dataset.tar')).as_posix()
    '/g/data/v10/some/dataset.tar'
    >>> normalise_nci_symlinks(Path('/g/data1a/v10/some/dataset.tar')).as_posix()
    '/g/data/v10/some/dataset.tar'
    >>> # Don't change other paths!
    >>> normalise_nci_symlinks(Path('/g/data/v10/some/dataset.tar')).as_posix()
    '/g/data/v10/some/dataset.tar'
    >>> normalise_nci_symlinks(Path('/Users/testuser/unrelated-path.yaml')).as_posix()
    '/Users/testuser/unrelated-path.yaml'
    """
    match = re.match(r"^/g/data[0-9a-z]+/(.*)", str(input_path))
    if not match:
        return input_path

    [offset] = match.groups()
    return Path("/g/data/" + offset)


def _dataset_name(ds_path: Path) -> str:
    """
    >>> _dataset_name(Path("example/LE07_L1GT_104078_20131209_20161119_01_T1.tar.gz"))
    'LE07_L1GT_104078_20131209_20161119_01_T1'
    >>> _dataset_name(Path("example/LE07_L1GT_104078_20131209_20161119_01_T1.tar"))
    'LE07_L1GT_104078_20131209_20161119_01_T1'
    >>> _dataset_name(Path("example/LE07_L1GT_104078_20131209_20161119_01_T2"))
    'LE07_L1GT_104078_20131209_20161119_01_T2'
    """
    # This is a little simpler than before :)
    return ds_path.stem.split(".")[0]


def _dataset_region_code(ds_path: Path) -> str:
    """
    >>> _dataset_region_code(Path("example/LE07_L1GT_104078_20131209_20161119_01_T1.tar.gz"))
    '104078'
    >>> _dataset_region_code(Path("example/LE07_L1GT_104078_20131209_20161119_01_T1.tar"))
    '104078'
    >>> _dataset_region_code(Path("example/LE07_L1GT_104078_20131209_20161119_01_T2"))
    '104078'
    """
    return _dataset_name(ds_path).split("_")[2]


if __name__ == "__main__":
    main()



"""
Prepare eo3 metadata for Sentinel-2 Level 1C data produced by Sinergise or ESA.

Takes ESA zipped datasets or Sinergise dataset directories
"""

import fnmatch
import json
import os.path
import re
import sys
import traceback
import uuid
import warnings
import zipfile
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, Optional, Tuple, Union

import click
import structlog
from attr import define
from datacube.config import LocalConfig
from datacube.index import index_connect
from datacube.model import Dataset
from datacube.ui.click import config_option, environment_option
from datacube.utils.uris import normalise_path
from defusedxml import minidom

from eodatasets3 import DatasetDoc, DatasetPrepare, names, serialise
from eodatasets3.prepare.s2_common import FolderInfo, RegionLookup
from eodatasets3.properties import Eo3Interface
from eodatasets3.ui import PathPath
from eodatasets3.utils import pass_config

# Static namespace to generate uuids for datacube indexing
S2_UUID_NAMESPACE = uuid.UUID("9df23adf-fc90-4ec7-9299-57bd536c7590")

_LOG = structlog.get_logger()

SENTINEL_MSI_BAND_ALIASES = {
    "01": "coastal_aerosol",
    "02": "blue",
    "03": "green",
    "04": "red",
    "05": "red_edge_1",
    "06": "red_edge_2",
    "07": "red_edge_3",
    "08": "nir_1",
    "8A": "nir_2",
    "09": "water_vapour",
    "10": "swir_1_cirrus",
    "11": "swir_2",
    "12": "swir_3",
}


def process_sinergise_product_info(product_path: Path) -> Dict:
    with product_path.open() as fp:
        product = json.load(fp)

    if len(product["tiles"]) > 1:
        raise NotImplementedError("No support for multi-tiled products yet")
    tile = product["tiles"][0]

    utm_zone = tile["utmZone"]
    latitude_band = tile["latitudeBand"]
    grid_square = tile["gridSquare"]
    return {
        "sentinel:product_name": product["name"],
        "sinergise_product_id": product["id"],
        "odc:region_code": f"{utm_zone}{latitude_band}{grid_square}",
        "sentinel:utm_zone": utm_zone,
        "sentinel:latitude_band": latitude_band,
        "sentinel:grid_square": grid_square,
    }


def _node_name(el):
    """Get a human-readable path to the given xml node"""
    branch = [el]
    while branch[-1].parentNode is not None:
        branch.append(branch[-1].parentNode)

    branch.reverse()
    return "/".join(n.localName for n in branch[1:])


def _value(root, *tags: str, type_=None):
    """Get the text contents of a tag in an xml document.

    Takes the tag name(s) to search for.
    """
    el = root

    for tag in tags:
        found = el.getElementsByTagName(tag)
        if not found:
            raise ValueError(f"Element not found in document: {tags!r}")
        if len(found) > 1:
            vs = "\n\t".join([_node_name(e) for e in found])
            raise ValueError(f"Multiple matches found for tag: {tags!r}: \n\t{vs}")

        el = found[0]

    el.normalize()
    if not len(el.childNodes) == 1:
        raise NotImplementedError(
            f"Not yet supported: Multiple child tags found for {tags!r}"
        )
    value = el.firstChild.data.strip()

    if not value:
        raise ValueError(f"Empty value for document at {tags!r}")

    if type_ is not None:
        value = type_(value)
    return value


def process_tile_metadata(contents: str) -> Dict:
    """
    Tile xml metadata format, as described by
    xmlns https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-1C_Tile_Metadata.xsd
    """
    root = minidom.parseString(contents)

    resolution = min(
        int(i.attributes["resolution"].value) for i in root.getElementsByTagName("Size")
    )
    tile_id = _value(root, "TILE_ID")

    region_code = tile_id.split("_")[-2]
    if not region_code.startswith("T"):
        raise RuntimeError(
            f"Tile id is not recognised -- not a region code? Please report this. {tile_id!r}"
        )
    region_code = region_code[1:]

    return {
        "datetime": _value(root, "SENSING_TIME"),
        "eo:cloud_cover": _value(root, "CLOUDY_PIXEL_PERCENTAGE", type_=float),
        "eo:gsd": resolution,
        "eo:sun_azimuth": _value(root, "Mean_Sun_Angle", "AZIMUTH_ANGLE", type_=float),
        "eo:sun_elevation": _value(root, "Mean_Sun_Angle", "ZENITH_ANGLE", type_=float),
        "odc:processing_datetime": _value(root, "ARCHIVING_TIME"),
        "sentinel:datastrip_id": _value(root, "DATASTRIP_ID"),
        "sentinel:sentinel_tile_id": tile_id,
        "odc:region_code": region_code,
    }


def process_datastrip_metadata(contents: str) -> Union[List[str], Dict]:
    """
    Datastrip metadata format, as described by
    xmlns https://psd-14.sentinel2.eo.esa.int/PSD/S2_PDI_Level-1C_Datastrip_Metadata.xsd

    Returns a list of tile names in this datastrip, and the properties that could be extracted.
    """
    root = minidom.parseString(contents)

    resolution_tags = root.getElementsByTagName("RESOLUTION")
    if resolution_tags:
        resolution = min(int(i.firstChild.data) for i in resolution_tags)
    elif root.getElementsByTagName("SPACECRAFT_NAME")[0].firstChild.data.startswith(
        "Sentinel-2"
    ):
        resolution = 10.0
    else:
        raise ValueError("No resolution in datastrip metadata and unknown craft")

    tile_ids = [
        tile_tag.attributes["tileId"] for tile_tag in root.getElementsByTagName("Tile")
    ]

    return tile_ids, {
        "sentinel:reception_station": _value(root, "RECEPTION_STATION"),
        "sentinel:processing_center": _value(root, "PROCESSING_CENTER"),
        "eo:gsd": resolution,
    }


def process_user_product_metadata(contents: str, filename_stem: str = None) -> Dict:
    root = minidom.parseString(contents)

    # - On newer datasets, get the product URI from metadata.
    # - On older datasets, get it from the filename.
    # (the metadata field is useless on older datasets, the filename is useless/fixed on newer datasets)
    raw_product_uri = _value(root, "PRODUCT_URI")

    if raw_product_uri.startswith("S2"):
        product_uri = raw_product_uri.split(".")[0]
    else:
        product_uri = filename_stem

    return {
        "eo:platform": _value(root, "SPACECRAFT_NAME"),
        "sat:relative_orbit": _value(root, "SENSING_ORBIT_NUMBER", type_=int),
        "sat:orbit_state": _value(root, "SENSING_ORBIT_DIRECTION").lower(),
        "sentinel:datatake_type": _value(root, "DATATAKE_TYPE"),
        "sentinel:processing_baseline": _value(root, "PROCESSING_BASELINE"),
        "sentinel:product_name": product_uri,
        "eo:cloud_cover": _value(root, "Cloud_Coverage_Assessment"),
    }


def _get_stable_id(p: Eo3Interface) -> uuid.UUID:
    # These should have been extracted from our metadata files!
    producer = p.producer
    product_name = p.properties["sentinel:product_name"]
    tile_id = p.properties["sentinel:sentinel_tile_id"]
    return uuid.uuid5(
        S2_UUID_NAMESPACE,
        ":".join((producer, product_name, tile_id)),
    )


def list_granules(dataset_location: Path) -> Optional[List[str]]:
    """
    Get a list of granule ids if it's a dataset that may contain multiple.
    """
    if dataset_location.suffix.lower() == ".json":
        # Sinergise
        return None

    # Each granule is a directory. Get a unique list of all the granule directory names (ie. IDs).
    granules = set()

    # We only want granule directories containing imagery files.
    # Some "null" datasets have been found with empty folders.
    granule_folder_pattern = re.compile(r"/GRANULE/([^/]+)/.*jp2$")

    with zipfile.ZipFile(dataset_location, "r") as z:
        for name in z.namelist():
            match = granule_folder_pattern.search(name)
            if match:
                granules.add(match.group(1))

    return sorted(granules)


def prepare_and_write(
    dataset_location: Path,
    output_yaml: Path,
    producer: str,
    granule_id: str = None,
    embed_location: bool = None,
) -> Tuple[DatasetDoc, Path]:
    if embed_location is None:
        # Default to embedding the location if they're not in the same folder.
        embed_location = output_yaml.parent not in dataset_location.parents
        _LOG.debug(
            "Auto-embed location?",
            auto_embed=bool(embed_location),
            data_location=dataset_location.parent,
            yaml_location=output_yaml.parent,
        )

    with DatasetPrepare(
        metadata_path=output_yaml,
        dataset_location=dataset_location,
    ) as p:
        p.properties["odc:producer"] = producer

        if producer == "esa.int":
            jp2_offsets = _extract_esa_fields(
                dataset_location, p, granule_id=granule_id
            )
        elif producer == "sinergise.com":
            jp2_offsets = _extract_sinergise_fields(dataset_location.parent, p)
        else:
            raise NotImplementedError(
                f"Unknown s2 producer {producer}. Expected 'sinergise.com' or 'esa.int'"
            )

        p.dataset_id = _get_stable_id(p)

        p.platform = _get_platform_name(p.properties)
        p.instrument = "MSI"
        p.constellation = "sentinel-2"

        # TODO: How to read collection number from metadata? (once ESA etc add one)
        collection_number = 0
        p.dataset_version = f"{collection_number}.0.{p.processed:%Y%m%d}"

        p.properties["odc:file_format"] = "JPEG2000"
        p.product_family = "level1"

        for path in jp2_offsets:
            band_number = _extract_band_number(path.stem)
            if band_number.lower() in ("tci", "pvi", "preview"):
                continue
            if band_number not in SENTINEL_MSI_BAND_ALIASES:
                raise RuntimeError(
                    f"Unknown band number {band_number!r} in image {path}"
                )

            p.note_measurement(
                path=path,
                name=SENTINEL_MSI_BAND_ALIASES[band_number],
                relative_to_dataset_location=True,
            )

        dataset_id, metadata_path = p.done(embed_location=embed_location)
        doc = serialise.from_doc(p.written_dataset_doc, skip_validation=True)
        if not doc.locations:
            doc.locations = [names.resolve_location(dataset_location)]
        return doc, metadata_path


def _get_platform_name(p: Mapping) -> str:
    # Grab it from the datastrip id,
    # Eg. S2B_OPER_MSI_L1C_DS_VGS4_20210426T010904_S20210425T235239_N03

    datastrip_id: str = p.get("sentinel:datastrip_id")
    if not datastrip_id:
        raise ValueError(
            "Could not find a sentinel datastrip ID after reading all metadata documents"
        )
    if not datastrip_id.lower().startswith("s2"):
        raise RuntimeError("Expected sentinel datastrip-id to start with 's2'!")
    platform_variant = datastrip_id[1:3].lower()
    platform_name = f"sentinel-{platform_variant}"
    return platform_name


def _extract_sinergise_fields(path: Path, p: DatasetPrepare) -> Iterable[Path]:
    """Extract Sinergise metadata and return list of image offsets"""
    product_info_path = path / "productInfo.json"
    metadata_xml_path = path / "metadata.xml"

    if not product_info_path.exists():
        raise ValueError(
            "No productInfo.json file found. "
            "Are you sure the input is a sinergise dataset folder?"
        )

    # Tile/Granule metadata
    p.properties.update(process_tile_metadata(metadata_xml_path.read_text()))
    p.note_accessory_file("metadata:s2_tile", metadata_xml_path)

    # Whole-product metadata
    for prop, value in process_sinergise_product_info(product_info_path).items():
        # We don't want to override properties that came from the (more-specific) tile metadata.
        if prop not in p.properties:
            p.properties[prop] = value
    p.note_accessory_file("metadata:sinergise_product_info", product_info_path)

    # TODO: sinergise folders could `process_datastrip_metadata()` in an outer directory?

    return list(path.glob("*.jp2"))


def _extract_esa_fields(
    dataset: Path, p: DatasetPrepare, granule_id: str
) -> Iterable[Path]:
    """Extract ESA metadata and return list of image offsets"""
    with zipfile.ZipFile(dataset, "r") as z:

        def find(*patterns) -> Iterable[str]:
            internal_folder_name = os.path.commonprefix(z.namelist())
            for s in z.namelist():
                if any(
                    re.match(pattern, s[len(internal_folder_name) :])
                    for pattern in patterns
                ):
                    yield s

        def one(*patterns) -> str:
            """Find one path ending in the given name"""

            [*matches] = find(*patterns)
            if len(matches) != 1:
                raise ValueError(
                    f"Expected exactly one file matching {patterns}, but found {len(matches)} of them in {dataset}:"
                    f"\n\t{matches}"
                )
            return matches[0]

        # If multiple datasetrips, find the one with this granule id?
        datastrip_files = find(r".*MTD_DS\.xml$", r"DATASTRIP/S2[^/]+/S2[^/]+\.xml$")
        if not datastrip_files:
            raise ValueError(f"No datastrip metadatas found in input? {dataset}")

        datastrip_md = None
        for datastrip_file in datastrip_files:
            inner_tiles, datastrip_md = process_datastrip_metadata(
                z.read(datastrip_file).decode("utf-8")
            )

            if granule_id in inner_tiles:
                # We found the datastrip for this tile.
                break

        if not datastrip_md:
            raise ValueError(
                f"None of the found datastrip metadatas cover this granule id {granule_id}"
            )

        p.properties.update(datastrip_md)
        p.note_accessory_file("metadata:s2_datastrip", datastrip_file)

        # Get the specific granule metadata
        [*tile_mds] = find(
            r".*MTD_TL\.xml$",
            rf"GRANULE/{granule_id}/S2.*\.xml" if granule_id else r"GRANULE/S2.*\.xml",
        )
        if not tile_mds:
            raise ValueError(
                "Could not find any tile metadata files in the dataset. "
                + (
                    f"(Searching for granule {granule_id!r})"
                    if granule_id
                    else "(Any granule)"
                )
            )
        if len(tile_mds) != 1:
            raise ValueError(
                f"Expected exactly one granule in package, since no granule id was provided: {dataset}:"
                f"\n\t{tile_mds}"
            )
        tile_md = tile_mds[0]

        p.properties.update(process_tile_metadata(z.read(tile_md).decode("utf-8")))
        p.note_accessory_file("metadata:s2_tile", tile_md)

        # Wider product metadata.
        user_product_md = one(r".*MTD_MSIL1C\.xml", r"S2.*\.xml")
        for prop, value in process_user_product_metadata(
            z.read(user_product_md).decode("utf-8"),
            filename_stem=Path(user_product_md).stem,
        ).items():
            # We don't want to override properties that came from the (more-specific) tile metadata.
            if prop not in p.properties:
                p.properties[prop] = value
        p.note_accessory_file("metadata:s2_user_product", user_product_md)

        # Get the list of images in the same tile folder.
        img_folder_offset = (Path(tile_md).parent / "IMG_DATA").as_posix()
        return [
            Path(p)
            for p in z.namelist()
            if p.startswith(img_folder_offset) and p.endswith(".jp2")
        ]


def _extract_band_number(name: str) -> str:
    """
    >>> _extract_band_number('B03')
    '03'
    >>> _extract_band_number('T55HFA_20201011T000249_B01')
    '01'
    """
    return name.split("_")[-1].replace("B", "")


def _rglob_with_self(path: Path, pattern: str) -> Iterable[Path]:
    if fnmatch.fnmatch(path.name, pattern):
        yield path
        return
    yield from path.rglob(pattern)


class YearMonth(click.ParamType):
    """A YYYY-MM string converted to an integer tuple"""

    name = "year-month"

    def convert(self, value, param, ctx):
        if value is None:
            return None
        hp = value.split("-")
        if len(hp) != 2:
            self.fail("Expect value in YYYY-MM format")

        year, month = hp
        if not year.isdigit() or not month.isdigit():
            self.fail("Expect value in YYYY-MM format")

        return int(year), int(month)


@define(order=True, hash=True)
class Job:
    """A dataset to process"""

    dataset_path: Path
    output_yaml_path: Path
    granule_id: Optional[str]

    # "sinergise.com" / "esa.int"
    producer: str
    embed_location: bool


@define(order=True, hash=True)
class InputDataset:
    name: str
    producer: str

    path: Path
    base_folder: Path

    @property
    def metadata(self) -> Optional[FolderInfo]:
        return FolderInfo.for_path(self.path)

    @property
    def granule_offsets(self) -> Optional[List[str]]:
        """
        Get the list of granule offsets in the dataset.

        Offsets are the subdirectory names used inside ESA dataset packages.

        Example: 'L1C_T55HFA_A018789_20201011T000244'
        """
        return list_granules(self.path)


REGION_CODE_FORMAT = re.compile(
    r"T"
    # A landsat path/row
    r"([0-9]{6}"
    # ... or a sentinel MGRS
    r"|[0-9A-Z]{5})"
)


def get_region_code_from_granule_offset(granule_id: str) -> str:
    """
    >>> get_region_code_from_granule_offset('L1C_T55HFA_A018789_20201011T000244')
    '55HFA'
    >>> get_region_code_from_granule_offset('S2A_OPER_MSI_L1C_TL_EPA__20161001T102755_A000900_T53KNU_N02.04')
    '53KNU'
    """
    for section in granule_id.split("_"):
        match = REGION_CODE_FORMAT.match(section)
        if match:
            return match.group(1)

    raise ValueError(f"Granule offset is in an unexpected form? {granule_id!r}")


@click.command(help=__doc__)
@click.option("-v", "--verbose", is_flag=True)
@click.argument(
    "datasets",
    type=PathPath(),
    nargs=-1,
)
@click.option(
    "-f",
    "--datasets-path",
    help="A file to read input dataset paths from, one per line",
    required=False,
    type=PathPath(
        exists=True, readable=True, dir_okay=False, file_okay=True, resolve_path=True
    ),
)
@click.option(
    "-j",
    "--jobs",
    "workers",
    help="Number of workers to run in parallel",
    type=int,
    default=1,
)
@click.option(
    "--overwrite-existing/--skip-existing",
    is_flag=True,
    default=False,
    help="Overwrite if exists (otherwise skip)",
)
@click.option(
    "--embed-location/--no-embed-location",
    is_flag=True,
    default=None,
    help="Embed the location of the dataset in the metadata? "
    "(if you wish to store them separately. default: auto)",
)
@click.option(
    "--always-granule-id/--never-granule-id",
    "always_granule_id",
    is_flag=True,
    default=None,
    help="Include the granule id in metadata filenames? (default: auto -- include only for multi-granule files). "
    "Beware that multi-granule datasets without a granule id in the filename will overwrite each-other",
)
@click.option(
    "--throughly-check-existing/--cheaply-check-existing",
    "thoroughly_check_existing",
    is_flag=True,
    default=False,
    help="Should we open every dataset to check if *all* inner granules have been produced? Default: false.",
)
@click.option(
    "--provider",
    default=None,
    type=click.Choice(
        [
            "sinergise.com",
            "esa.int",
        ]
    ),
    help="Restrict scanning to only packages of the given provider. "
    "(ESA assumes a zip file, sinergise a directory)",
)
@click.option(
    "--output-base",
    help="Write metadata files into a directory instead of alongside each dataset",
    required=False,
    type=PathPath(
        exists=True, writable=True, dir_okay=True, file_okay=False, resolve_path=True
    ),
)
@click.option(
    "--input-relative-to",
    help="Input root folder that should be used for the subfolder hierarchy in the output-base",
    required=False,
    type=PathPath(
        exists=True, writable=True, dir_okay=True, file_okay=False, resolve_path=True
    ),
)
@click.option(
    "--only-regions-in-file",
    help="Only process datasets in the given regions. Expects a file with one region code per line. "
    "(Note that some older ESA datasets have no region code, and will not match any region here.)",
    required=False,
    type=PathPath(
        exists=True, readable=True, dir_okay=False, file_okay=True, resolve_path=True
    ),
)
@click.option(
    "--after-month",
    help="Limit the scan to datasets newer than a given month "
    "(expressed as {year}-{month}, eg '2010-01')",
    required=False,
    type=YearMonth(),
)
@click.option(
    "--before-month",
    help="Limit the scan to datasets older than the given month "
    "(expressed as {year}-{month}, eg '2010-01')",
    required=False,
    type=YearMonth(),
)
@click.option(
    "--index",
    "index_to_odc",
    is_flag=True,
    default=False,
    help="Index newly-generated metadata into the configured datacube",
)
@click.option(
    "--dry-run",
    help="Show what would be created, but don't create anything",
    is_flag=True,
    default=False,
)
@environment_option
@config_option
@pass_config(required=False)
def main(
    local_config: LocalConfig,
    output_base: Optional[Path],
    input_relative_to: Optional[Path],
    datasets: Tuple[Path],
    datasets_path: Optional[Path],
    provider: Optional[str],
    overwrite_existing: bool,
    verbose: bool,
    workers: int,
    thoroughly_check_existing: bool,
    embed_location: Optional[bool],
    only_regions_in_file: Optional[Path],
    before_month: Optional[Tuple[int, int]],
    after_month: Optional[Tuple[int, int]],
    dry_run: bool,
    always_granule_id: Optional[bool],
    index_to_odc: bool,
):
    if sys.argv[1] == "sentinel-l1c":
        warnings.warn(
            "Command name 'sentinel-l1c-prepare' is deprecated: remove the 'c', and use `sentinel-l1-prepare`"
        )

    included_regions = None
    if only_regions_in_file:
        included_regions = set(only_regions_in_file.read_text().splitlines())

    if datasets_path:
        datasets = [
            *datasets,
            *(
                normalise_path(p.strip())
                for p in (datasets_path.read_text().splitlines())
            ),
        ]

    _LOG.info("kickoff", path_count=len(datasets), worker_count=workers)

    # Are we indexing on success?
    index = None
    if index_to_odc:
        _LOG.info("Indexing new datasets", local_config=local_config)
        index = index_connect(local_config, application_name="s2-prepare")
        products = {}

        def on_success(dataset: DatasetDoc, dataset_path: Path):
            """
            Index the dataset
            """
            product_name = dataset.product.name
            product = products.get(product_name)
            if not product:
                product = index.products.get_by_name(product_name)
                if not product:
                    raise ValueError(f"Product {product_name} not found in ODC index")
                products[product_name] = product

            index.datasets.add(
                Dataset(product, serialise.to_doc(dataset), uris=dataset.locations)
            )
            _LOG.debug(
                "Indexed dataset", dataset_id=dataset.id, dataset_path=dataset_path
            )

    else:

        def on_success(dataset: DatasetDoc, dataset_path: Path):
            """Nothing extra"""

    def find_inputs_in_path(input_path: Path) -> Iterable[InputDataset]:
        """
        Scan the input path for our key identifying files of a package.
        """
        found_something = False
        if provider == "sinergise.com" or not provider:
            for p in _rglob_with_self(input_path, "tileInfo.json"):
                found_something = True
                yield InputDataset(
                    producer="sinergise.com",
                    # Dataset location is the metadata file itself.
                    path=p,
                    # Output is a sibling metadata file, with the same name as the folder (usually S2A....).
                    base_folder=p.parent.parent,
                    name=p.parent.stem,
                )
        if provider == "esa.int" or not provider:
            for p in _rglob_with_self(input_path, "*.zip"):
                found_something = True
                yield InputDataset(
                    producer="esa.int",
                    # Dataset location is the zip file
                    path=p,
                    # Metadata is a sibling file with a metadata suffix.
                    base_folder=p.parent,
                    name=p.stem,
                )
        if not found_something:
            raise ValueError(
                f"No S2 datasets found in given path {input_path}. "
                f"Expected either Sinergise (productInfo.json) files or ESA zip files to be contained in it."
            )

    def find_jobs() -> Iterable[Job]:
        region_lookup = RegionLookup()

        nonlocal input_relative_to, embed_location
        for input_path in datasets:
            first = True
            for found_dataset in find_inputs_in_path(input_path):
                _LOG.debug("found_dataset", name=found_dataset.name)
                # Make sure we tick progress on extra datasets that were found.
                if not first:
                    first = False

                # Filter based on metadata
                info = found_dataset.metadata

                # Skip regions that are not in the limit?
                if included_regions or before_month or after_month:
                    if info is None:
                        raise ValueError(
                            f"Cannot filter from non-standard folder layout: {found_dataset.path} "
                            f" expected of form L1C/yyyy/yyyy-mm/area/S2_.."
                        )

                    if included_regions:
                        # If it's an older dataset without a region, try to map its area to a known region.
                        if info.region_code is None:
                            for region in region_lookup.get(info.area):
                                if region in included_regions:
                                    _LOG.debug(
                                        "mapped_area_match",
                                        input_area=info.area,
                                        region_match=region,
                                    )
                                    break
                            else:
                                _LOG.debug(
                                    "skipping.mapped_area_not_in_regions",
                                    input_area=info.area,
                                )
                                continue
                        elif info.region_code not in included_regions:
                            _LOG.debug(
                                "skipping.region_not_in_region_list",
                                region_code=info.region_code,
                            )
                            continue

                    if after_month is not None:
                        year, month = after_month

                        if info.year < year or (
                            info.year == year and info.month < month
                        ):
                            _LOG.debug(
                                "skipping.too_old",
                                dataset_year_month=(info.year, info.month),
                                max_year_month=(year, month),
                            )
                            continue
                    if before_month is not None:
                        year, month = before_month

                        if info.year > year or (
                            info.year == year and info.month > month
                        ):
                            _LOG.debug(
                                "skipping.too_young",
                                dataset_year_month=(info.year, info.month),
                                min_year_month=(year, month),
                            )
                            continue

                # Put outputs in a different folder?
                if output_base:
                    # What base folder should we choose for creating subfolders in the output?
                    if input_relative_to is None:
                        input_relative_to = _get_default_relative_folder_base(
                            found_dataset.base_folder
                        )

                    output_folder = output_base / found_dataset.base_folder.relative_to(
                        input_relative_to
                    )
                    # Default to true.
                    if embed_location is None:
                        embed_location = True
                else:
                    output_folder = found_dataset.base_folder
                    # Default to false
                    if embed_location is None:
                        embed_location = False

                # It's very slow to read the list of inner granules.
                #
                # So, if we're not thoroughly checking for missing outputs.
                if (
                    (not thoroughly_check_existing)
                    # ... and any outputs exist at all
                    and list(
                        output_folder.glob(f"{found_dataset.name}*.odc-metadata.yaml")
                    )
                    # ... and we're not overwriting our outputs
                    and not overwrite_existing
                ):
                    # Skip it!
                    _LOG.debug(
                        "At least one output exists: skipping.",
                        dataset_name=found_dataset.name,
                    )
                    continue

                # This has to read the files, so can be slow. That's why we try to skip above if possible.
                granule_offsets = found_dataset.granule_offsets

                # When granule_id is None, it means process all without filtering.
                if not granule_offsets:
                    granule_offsets = [None]
                else:
                    _LOG.debug("found_granules", granule_count=len(granule_offsets))

                for granule_offset in granule_offsets:
                    # Now that we've extracted actual granules, try again to filter by region.
                    if granule_offset and included_regions:
                        region_code = get_region_code_from_granule_offset(
                            granule_offset
                        )
                        if region_code not in included_regions:
                            _LOG.debug(
                                "skipping.granule_out_of_region",
                                granule_offset=granule_offset,
                                region_code=region_code,
                            )
                            continue

                    if always_granule_id or (
                        # None means 'auto': ie. automatically include granule id when there are multiple granules
                        always_granule_id is None and len(granule_offsets) > 1
                    ):
                        yaml_filename = (
                            f"{found_dataset.name}.{granule_offset}.odc-metadata.yaml"
                        )
                    else:
                        yaml_filename = f"{found_dataset.name}.odc-metadata.yaml"

                    output_yaml = output_folder / yaml_filename
                    if output_yaml.exists():
                        if not overwrite_existing:
                            _LOG.debug(
                                "Output exists: skipping.", output_yaml=output_yaml
                            )
                            continue

                        _LOG.debug(
                            "Output exists: overwriting.", output_yaml=output_yaml
                        )

                    _LOG.info(
                        "queued",
                        dataset_name=found_dataset.name,
                        granule=granule_offset or "any",
                    )
                    yield Job(
                        dataset_path=found_dataset.path,
                        output_yaml_path=output_yaml,
                        producer=found_dataset.producer,
                        granule_id=granule_offset,
                        embed_location=embed_location,
                    )

    errors = 0

    if dry_run:
        _LOG.info("Dry run: not writing any files.")

    # If only one process, call it directly.
    # (Multiprocessing makes debugging harder, so we prefer to make it optional)
    successes = 0

    try:
        if workers == 1 or dry_run:
            for job in find_jobs():
                try:
                    if dry_run:
                        _LOG.info(
                            "Would write dataset",
                            dataset_path=job.dataset_path,
                            output_yaml_path=job.output_yaml_path,
                        )
                    else:
                        file_already_existed = (
                            overwrite_existing and job.output_yaml_path.exists()
                        )

                        dataset, path = prepare_and_write(
                            job.dataset_path,
                            job.output_yaml_path,
                            job.producer,
                            granule_id=job.granule_id,
                            embed_location=job.embed_location,
                        )
                        _LOG.info(
                            "Wrote dataset",
                            dataset_id=dataset.id,
                            dataset_path=path,
                            output_yaml=job.output_yaml_path,
                        )

                        try:
                            on_success(dataset, path)
                        except Exception:
                            # If the post-processing function fails, we don't want this "unfinished" file still around.
                            # We'd rather let it be re-created next time.
                            if not file_already_existed:
                                _LOG.info(
                                    "Cleaning new file due to indexing failure.",
                                    output_yaml=job.output_yaml_path,
                                )
                                job.output_yaml_path.unlink()

                            raise

                    successes += 1
                except Exception:
                    _LOG.exception("failed_job", job=job)
                    errors += 1
        else:
            with Pool(processes=workers) as pool:
                for res in pool.imap_unordered(_write_dataset_safe, find_jobs()):
                    if isinstance(res, str):
                        _LOG.error(res)
                        errors += 1
                    else:
                        dataset, path = res
                        _LOG.info(
                            "Wrote dataset", dataset_id=dataset.id, dataset_path=path
                        )
                        on_success(dataset, path)
                        successes += 1
                pool.close()
                pool.join()
    finally:
        if index is not None:
            index.close()

    _LOG.info("completed", success_count=successes, failure_count=errors)
    sys.exit(errors)


def _get_default_relative_folder_base(path: Path) -> Optional[Path]:
    for parent in path.parents:
        if parent.name.lower() == "l1c":
            input_relative_to = parent
            _LOG.debug("found_base_folder", inputs_are_relative_to=input_relative_to)
            break
    else:
        raise ValueError(
            "Unknown root folder for path subfolders. "
            "(Hint: specify --input-relative-to with a parent folder of the inputs. "
            "Outputs will use the same subfolder structure.)"
        )

    return input_relative_to


def _write_dataset_safe(job: Job) -> Union[Tuple[DatasetDoc, Path], str]:
    """
    A wrapper around `prepare_and_write` that catches exceptions and makes them
    serialisable as error strings.

    (for use in multiprocessing pools etc.)
    """
    try:
        dataset, path = prepare_and_write(
            job.dataset_path,
            job.output_yaml_path,
            job.producer,
            granule_id=job.granule_id,
            embed_location=job.embed_location,
        )
        return dataset, path
    except Exception:
        return f"Failed to write dataset: {job}\n" + traceback.format_exc()


if __name__ == "__main__":
    main()



"""
Package WAGL HDF5 Outputs

This converts the HDF5 file (and sibling fmask/gqa files) into
GeoTIFFS (COGs) with datacube metadata using the DEA naming conventions
for files.
"""

import contextlib
import os
import re
import sys
import zipfile
from datetime import datetime, timedelta
from enum import Enum
from math import isnan
from os.path import join
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple
from uuid import UUID

import attr
import defusedxml.ElementTree as Et
import fiona
import numpy
import rasterio
from affine import Affine
from boltons.iterutils import PathAccessError, get_path
from click import secho
from rasterio import DatasetReader
from rasterio.crs import CRS
from rasterio.enums import Resampling

from eodatasets3 import DatasetAssembler, images, serialise, utils
from eodatasets3.images import GridSpec
from eodatasets3.model import DatasetDoc
from eodatasets3.properties import Eo3Interface
from eodatasets3.serialise import loads_yaml
from eodatasets3.ui import bool_style
from eodatasets3.utils import default_utc, flatten_dict

try:
    import h5py
except ImportError:
    sys.stderr.write(
        "eodatasets3 has not been installed with the wagl extras. \n"
        "    Try `pip install eodatasets3[wagl]\n"
    )
    raise

POSSIBLE_PRODUCTS = ("nbar", "nbart", "lambertian", "sbt")
DEFAULT_PRODUCTS = ("nbar", "nbart")

_THUMBNAILS = {
    "nbar": ("nbar:red", "nbar:green", "nbar:blue"),
    "nbart": ("nbart:red", "nbart:green", "nbart:blue"),
}

os.environ["CPL_ZIP_ENCODING"] = "UTF-8"

FILENAME_TIF_BAND = re.compile(
    r"(?P<prefix>(?:.*_)?)(?P<band_name>B[0-9][A0-9]|B[0-9]*|B[0-9a-zA-z]*)"
    r"(?P<extension>\....)"
)
PRODUCT_SUITE_FROM_GRANULE = re.compile("(L1[GTPCS]{1,2})")

# Map wagl dataset band IDs to ESA MTD.xml Band IDs
WAGL_TO_ESA_BAND_NUMBER = {
    "1": "0",
    "2": "1",
    "3": "2",
    "4": "3",
    "5": "4",
    "6": "5",
    "7": "6",
    "8": "7",
    "8A": "8",
    "9": "9",
    "10": "10",
    "11": "11",
    "12": "12",
}


class ProductMaturity(Enum):
    provisional = "provisional"
    stable = "stable"


# a dictionary of band_id: (mask_type, mask_uri)
#     (the mask URI is a string loadable by fiona. Could be a URL or file path.)
BandMasks = Dict[str, Tuple[str, str]]


def _find_h5_paths(h5_obj: h5py.Group, dataset_class: str = "") -> List[str]:
    """
    Find all objects in a h5 of the given class, returning their path.

    (class examples: IMAGE, TABLE. SCALAR)
    """
    items = []

    def _find(name, obj):
        if obj.attrs.get("CLASS") == dataset_class:
            items.append(name)

    h5_obj.visititems(_find)
    return items


def _unpack_products(
    p: DatasetAssembler,
    product_list: Iterable[str],
    h5group: h5py.Group,
    granule: "Granule",
) -> None:
    """
    Unpack and package the NBAR and NBART products.
    """
    # listing of all datasets of IMAGE CLASS type
    img_paths = _find_h5_paths(h5group, "IMAGE")

    for product in product_list:
        with sub_product(product, p):
            for pathname in [p for p in img_paths if f"/{product.upper()}/" in p]:
                with do(f"Path {pathname!r}"):
                    dataset = h5group[pathname]
                    band_masks = get_quality_masks(dataset, granule)
                    band_name = utils.normalise_band_name(dataset.attrs["alias"])
                    write_measurement_h5(
                        p,
                        f"{product}:{band_name}",
                        dataset,
                        band_masks=band_masks,
                        overview_resampling=Resampling.average,
                        file_id=_file_id(dataset),
                    )

            if product in _THUMBNAILS:
                red, green, blue = _THUMBNAILS[product]
                with do(f"Thumbnailing {product}"):
                    p.write_thumbnail(
                        red,
                        green,
                        blue,
                        static_stretch=(1, 3000),
                        # Because of our strange sub-products and filename standards, we want the
                        # 'kind' to be included in the recorded thumbnail accessory metadata,
                        # but not in the filename.
                        # So we manually calculate a filename without the 'kind' field included.
                        kind=product,
                        path=p.names.thumbnail_filename(),
                    )


def get_quality_masks(dataset: h5py.Dataset, granule: "Granule") -> BandMasks:
    """
    Get a dictionary of any quality masks to apply for each band (if any).
    """

    if dataset.attrs["band_id"] not in WAGL_TO_ESA_BAND_NUMBER.keys():
        return {}

    platform_name = granule.wagl_metadata["source_datasets"]["platform_id"].lower()
    if "sentinel" not in platform_name:
        return {}

    level1_data_path = granule.source_level1_data

    granule_metadata_xml = Path(
        granule.source_level1_metadata.accessories["metadata:s2_tile"].path
    )

    if not level1_data_path.exists():
        raise ValueError(f"Input level 1 data not found {level1_data_path}")

    if level1_data_path.suffix == ".zip":
        product_root = granule_metadata_xml.parts[0]
        # Parse tile metadata xml file
        with zipfile.ZipFile(level1_data_path, "r") as z:
            mtd_dict = raw_mask_offsets_from_granule_xml(
                z.read(granule_metadata_xml.as_posix())
            )
            masks: BandMasks = {}
            # Convert the raw mask offsets to loadable fiona paths.
            for band_id in list(mtd_dict.keys()):
                type_, location = mtd_dict[band_id]
                mask_offset = join(product_root, location)
                if mask_offset not in z.namelist():
                    mask_offset = str(
                        granule_metadata_xml.parent / "QI_DATA" / location
                    )
                info = z.getinfo(mask_offset)
                # Mask under 500 bytes are empty, which kills fiona on load.
                # (From scanning, non-empty ones in our collection are all far larger than 1000bytes,
                #  so this doesn't need precision)
                if info.file_size < 500:
                    continue
                masks[band_id] = (
                    type_,
                    f"zip+file://{level1_data_path.as_posix()}!/{mask_offset}",
                )
        return masks
    # Sinergise data comes as a directory.
    elif level1_data_path.is_dir():
        mtd_dict = raw_mask_offsets_from_granule_xml(
            (level1_data_path / granule_metadata_xml).read_bytes()
        )
        masks: BandMasks = {}
        for band_id in list(mtd_dict.keys()):
            type_, location = mtd_dict[band_id]
            mask_path = level1_data_path / location
            # Sinergise use the original ESA metadata document, but have a new directory structure
            # So the metadata mask locations are (always?) wrong.
            # (It's in a 'qi' folder with the original filename)
            if not mask_path.exists():
                mask_path = level1_data_path / "qi" / Path(location).name

            # Skip small ones. See reasoning in ESA block above.
            if mask_path.stat().st_size < 500:
                continue
            masks[band_id] = (type_, mask_path.as_posix())
        return masks
    else:
        raise RuntimeError(
            "Input level 1 data must be a zip file (ESA) or directory (Sinergise)"
        )


def raw_mask_offsets_from_granule_xml(granule_metadata_xml: bytes):
    root = Et.fromstring(granule_metadata_xml, forbid_dtd=True)

    return {
        # Append metadata bandId as key and mask type and path as tuple vals to dict
        e.attrib["bandId"]: (e.attrib["type"], e.text)
        for e in root.findall(".//Pixel_Level_QI/MASK_FILENAME")
        if "bandId" in e.attrib
        and "type" in e.attrib
        and e.attrib["type"] in ("MSK_QUALIT", "MSK_TECQUA")
    }


def load_and_mask_data(g: h5py.Dataset, masks: BandMasks):
    band_id = g.attrs["band_id"]
    esa_band = WAGL_TO_ESA_BAND_NUMBER[band_id]
    if esa_band not in masks:
        # No mask needs to be applied.
        data_array = g[:] if hasattr(g, "chunks") else g
        return data_array

    type_, mask_string_path = masks[esa_band]
    if type_ == "MSK_TECQUA":
        return mask_h5_vector(
            g,
            mask_string_path,
        )
    elif type_ == "MSK_QUALIT":
        return mask_h5_raster(
            g,
            mask_string_path,
        )
    else:
        raise ValueError(f"unknown mask type {type_}")


def mask_h5_vector(
    dataset: h5py.Dataset,
    mask_string_path,
):
    """
    Mask a hdf5 dataset using the gml files provided in the path.
    """
    data_array = dataset[:] if hasattr(dataset, "chunks") else dataset
    # Open the gml file
    with fiona.open(mask_string_path) as gml:
        shapes = [feature["geometry"] for feature in gml]
        if not shapes:
            return data_array
        mask_array = rasterio.features.rasterize(
            shapes,
            out_shape=data_array.shape,
            fill=0,
            transform=Affine.from_gdal(*dataset.attrs["geotransform"]),
        )
        return numpy.where(mask_array == 0, data_array, dataset.attrs["no_data_value"])


def mask_h5_raster(
    dataset: h5py.Dataset,
    mask_string_path,
):
    """
    Mask a hdf5 dataset using the gml files provided in the path.
    """
    data_array = dataset[:] if hasattr(dataset, "chunks") else dataset

    # Open the .jp2 file
    with rasterio.open(mask_string_path) as source_ds:
        # Mask layer info:
        # https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-2-msi/level-1c/masks
        # MSI lost data layer
        source_layer1 = source_ds.read(3)
        # MSI degraded data layer
        source_layer2 = source_ds.read(4)
        # Mask where missing OR degraded packets
        source_layer = source_layer1 | source_layer2
        masked_data = numpy.where(
            source_layer == 0, data_array, dataset.attrs["no_data_value"]
        )
        return masked_data


def write_measurement_h5(
    p: DatasetAssembler,
    full_name: str,
    g: h5py.Dataset,
    band_masks: Optional[BandMasks] = None,
    overviews=images.DEFAULT_OVERVIEWS,
    overview_resampling=Resampling.nearest,
    expand_valid_data=True,
    file_id: str = None,
):
    """
    Write a measurement by copying it from a hdf5 dataset.
    """
    if band_masks:
        data = load_and_mask_data(g, band_masks)
    else:
        data = g[:] if hasattr(g, "chunks") else g

    product_name, band_name = full_name.split(":")
    p.write_measurement_numpy(
        array=data,
        grid_spec=images.GridSpec(
            shape=g.shape,
            transform=Affine.from_gdal(*g.attrs["geotransform"]),
            crs=CRS.from_wkt(g.attrs["crs_wkt"]),
        ),
        nodata=g.attrs.get("no_data_value"),
        overviews=overviews,
        overview_resampling=overview_resampling,
        expand_valid_data=expand_valid_data,
        file_id=file_id,
        # Because of our strange sub-products and filename standards, we want the
        # product_name to be included in the recorded band metadata,
        # but not in its filename.
        # So we manually calculate a filename without the extra product name prefix.
        name=full_name,
        path=p.names.measurement_filename(band_name, "tif", file_id=file_id),
    )


def _file_id(dataset: h5py.Dataset) -> str:
    """
    Devise a file id for the given dataset (using its attributes)

    Eg. 'band01'
    """
    # What we have to work with:
    # >>> print(repr((dataset.attrs["band_id"], dataset.attrs["band_name"], dataset.attrs["alias"])))
    # ('1', 'BAND-1', 'Blue')

    band_name = dataset.attrs["band_id"]

    # A purely numeric id needs to be formatted 'band01' according to naming conventions.
    return utils.normalise_band_name(band_name)


def _unpack_observation_attributes(
    p: DatasetAssembler,
    res_grp: h5py.Group,
):
    """
    Unpack the angles + other supplementary datasets produced by wagl.
    Currently only the mode resolution group gets extracted.
    """

    def _write(section: str, dataset_names: Sequence[str]):
        """
        Write supplementary attributes as measurement.
        """
        for dataset_name in dataset_names:
            o = f"{section}/{dataset_name}"
            with do(f"Path {o!r} "):
                measurement_name = utils.normalise_band_name(dataset_name)
                write_measurement_h5(
                    p,
                    f"oa:{measurement_name}",
                    res_grp[o],
                    # We only use the product bands for valid data calc, not supplementary.
                    # According to Josh: Supplementary pixels outside of the product bounds are implicitly invalid.
                    expand_valid_data=False,
                    overviews=None,
                )

    _write(
        "SATELLITE-SOLAR",
        [
            "SATELLITE-VIEW",
            "SATELLITE-AZIMUTH",
            "SOLAR-ZENITH",
            "SOLAR-AZIMUTH",
            "RELATIVE-AZIMUTH",
            "TIME-DELTA",
        ],
    )
    _write("INCIDENT-ANGLES", ["INCIDENT-ANGLE", "AZIMUTHAL-INCIDENT"])
    _write("EXITING-ANGLES", ["EXITING-ANGLE", "AZIMUTHAL-EXITING"])
    _write("RELATIVE-SLOPE", ["RELATIVE-SLOPE"])
    _write("SHADOW-MASKS", ["COMBINED-TERRAIN-SHADOW"])


def get_oa_resolution_group(
    resolution_groups: Dict[tuple, h5py.Group],
    platform: str,
    oa_resolution: Optional[Tuple[float, float]],
) -> h5py.Group:
    # None specified? Figure out a default.

    if oa_resolution is None:
        # For Landsat, we only cared about packaging OA data for the "common"
        # bands (not panchromatic). So we always pick the higher resolution.
        if platform.startswith("landsat"):
            oa_resolution = max(resolution_groups.keys())
        elif platform.startswith("sentinel"):
            oa_resolution = (20.0, 20.0)
        else:
            raise NotImplementedError(
                f"Don't know how to choose a default OA resolution for platform {platform !r}"
            )

    res_grp = resolution_groups.get(oa_resolution)
    if res_grp is None:
        raise RuntimeError(
            f"Resolution {oa_resolution} not found in input. "
            f"Have resolutions {tuple(resolution_groups.keys())}"
        )

    return res_grp


def _create_contiguity(
    p: DatasetAssembler,
    product_list: Iterable[str],
    resolution_yx: Tuple[float, float],
    timedelta_product: str = "nbar",
    timedelta_data: numpy.ndarray = None,
):
    """
    Create the contiguity (all pixels valid) dataset.

    Write a contiguity mask file based on the intersection of valid data pixels across all
    bands from the input files.
    """
    for product in product_list:
        contiguity = None
        # A contiguity layer for each product
        for grid, band_name, path in p.iter_measurement_paths():
            if not band_name.startswith(f"{product.lower()}:"):
                continue
            # Only our given res group (no pan band in Landsat)
            if grid.resolution_yx != resolution_yx:
                continue
            with rasterio.open(path) as ds:
                ds: DatasetReader
                out_shape = ds.shape
                contiguity = numpy.ones(out_shape, dtype="uint8")
                geobox = GridSpec.from_rio(ds)
                break

        if contiguity is None:
            raise ValueError(f"no matching band with resolution {resolution_yx}")

        for grid, band_name, path in p.iter_measurement_paths():
            if not band_name.startswith(f"{product.lower()}:"):
                continue
            # Only our given res group (no pan band in Landsat)
            if (p.platform.startswith("landsat")) and (
                grid.resolution_yx != resolution_yx
            ):
                continue
            with rasterio.open(path) as ds:
                ds: DatasetReader
                contiguity &= (
                    ds.read(
                        ds.count, out_shape=out_shape, resampling=Resampling.nearest
                    )
                    > 0
                )

        p.write_measurement_numpy(
            f"oa:{product.lower()}_contiguity",
            contiguity,
            geobox,
            nodata=255,
            overviews=None,
            expand_valid_data=False,
            # Because of our strange sub-products and filename standards, we want the
            # 'oa_' prefix to be included in the recorded band metadata,
            # but not in its filename.
            # So we manually calculate a filename without the extra prefix.
            path=p.names.measurement_filename(f"{product.lower()}_contiguity"),
        )

        # masking the timedelta_data with contiguity mask to get max and min timedelta within the NBAR product
        # footprint for Landsat sensor. For Sentinel sensor, it inherits from level 1 yaml file
        if timedelta_data is not None and product.lower() == timedelta_product:
            valid_timedelta_data = numpy.ma.masked_where(
                contiguity == 0, timedelta_data
            )

            def offset_from_center(v: numpy.datetime64):
                return p.datetime + timedelta(
                    microseconds=v.astype(float) * 1_000_000.0
                )

            p.datetime_range = (
                offset_from_center(numpy.ma.min(valid_timedelta_data)),
                offset_from_center(numpy.ma.max(valid_timedelta_data)),
            )


@contextlib.contextmanager
def do(name: str, heading=False, **fields):
    """
    Informational logging.

    TODO: move this to the cli. It shouldn't be part of library usage.
    """
    single_line = not heading

    def val(v: Any):
        if isinstance(v, bool):
            return bool_style(v)
        if isinstance(v, Path):
            return repr(str(v))
        return repr(v)

    if heading:
        name = f"\n{name}"
    fields = " ".join(f"{k}:{val(v)}" for k, v in fields.items())
    secho(f"{name} {fields} ", nl=not single_line, fg="blue" if heading else None)
    yield
    if single_line:
        secho("(done)")


@contextlib.contextmanager
def sub_product(name: str, p: Eo3Interface):
    """
    Set the product family temporarily within a block of code.

    This is done for sub-products that WAGL contains, which have
    a different 'family' in their filenames.
    """
    with do(f"Product {name}", heading=True):
        original_family = p.product_family
        # We delete first to show that we're deliberately changing the value (no 'overridding property" warning)
        del p.product_family
        p.product_family = name
        yield
        del p.product_family
        p.product_family = original_family


def _extract_reference_code(p: DatasetAssembler, granule: str) -> Optional[str]:
    matches = None
    if p.platform.startswith("landsat"):
        matches = re.match(r"L\w\d(?P<reference_code>\d{6}).*", granule)
    elif p.platform.startswith("sentinel-2"):
        matches = re.match(r".*_T(?P<reference_code>\d{1,2}[A-Z]{3})_.*", granule)

    if matches:
        [reference_code] = matches.groups()
        # TODO name properly
        return reference_code
    return None


@attr.s(auto_attribs=True)
class Granule:
    """
    A single granule in a hdf5 file, with optional corresponding fmask/gqa/etc files.

    You probably want to make one by using `Granule.for_path()`
    """

    name: str
    wagl_hdf5: Path
    wagl_metadata: Dict
    source_level1_metadata: Optional[DatasetDoc]
    source_level1_data: Optional[Path] = None

    fmask_doc: Optional[Dict] = None
    fmask_image: Optional[Path] = None
    s2cloudless_prob: Optional[Path] = None
    s2cloudless_mask: Optional[Path] = None
    s2cloudless_doc: Optional[Dict] = None
    gqa_doc: Optional[Dict] = None
    tesp_doc: Optional[Dict] = None

    @classmethod
    def for_path(
        cls,
        wagl_hdf5: Path,
        granule_names: Optional[Sequence[str]] = None,
        level1_metadata_path: Optional[Path] = None,
        level1_data_path: Optional[Path] = None,
        fmask_image_path: Optional[Path] = None,
        fmask_doc_path: Optional[Path] = None,
        s2cloudless_prob_path: Optional[Path] = None,
        s2cloudless_mask_path: Optional[Path] = None,
        s2cloudless_doc_path: Optional[Path] = None,
        gqa_doc_path: Optional[Path] = None,
        tesp_doc_path: Optional[Path] = None,
        allow_missing_provenance: bool = False,
    ):
        """
        Create granules by scanning the given hdf5 file.

        Optionally specify additional files and level1 path.

        If they are not specified it look for them using WAGL's output naming conventions.
        :param allow_missing_provenance:
        """
        if not wagl_hdf5.exists():
            raise ValueError(f"Input hdf5 doesn't exist {wagl_hdf5}")

        with h5py.File(wagl_hdf5, "r") as fid:
            granule_names = granule_names or fid.keys()

            for granule_name in granule_names:
                if granule_name not in fid:
                    raise ValueError(
                        f"Granule {granule_name!r} not found in file {wagl_hdf5}"
                    )

                wagl_doc_field = get_path(fid, (granule_name, "METADATA", "CURRENT"))
                if not wagl_doc_field:
                    raise ValueError(
                        f"Granule contains no wagl metadata: {granule_name} in {wagl_hdf5}"
                    )

                [wagl_doc] = loads_yaml(wagl_doc_field[()])

                # If they don't specify a level1, default to the one in the wagl doc.
                if not level1_data_path:
                    level1_data_path = Path(
                        get_path(wagl_doc, ("source_datasets", "source_level1"))
                    )
                level1_doc = _load_level1_doc(
                    level1_data_path, level1_metadata_path, allow_missing_provenance
                )

                fmask_image_path = fmask_image_path or wagl_hdf5.with_name(
                    f"{granule_name}.fmask.img"
                )
                if not fmask_image_path.exists():
                    raise ValueError(f"No fmask image found at {fmask_image_path}")

                fmask_doc_path = fmask_doc_path or fmask_image_path.with_suffix(".yaml")
                if not fmask_doc_path.exists():
                    raise ValueError(f"No fmask found at {fmask_doc_path}")
                with fmask_doc_path.open("r") as fl:
                    [fmask_doc] = loads_yaml(fl)

                if "sentinel" in wagl_doc["source_datasets"]["platform_id"].lower():
                    # We need the level1 data for S2 in order to apply quality masks.
                    if not level1_data_path.exists():
                        # TODO: We could infer this from the 'location:' field of the metadata, when they
                        #       have embedded locations
                        # If the data directory is a sibling, we can infer it.
                        match = METADATA_DOC_STEM.match(level1_metadata_path.name)
                        if not match:
                            raise ValueError(
                                f"Can't infer level1 data path from {level1_data_path}"
                            )

                        # For ESA: It may be at an identical path, with zip suffix instead of *.odc-metadata.yaml
                        level1_data_path = level1_metadata_path.parent / (
                            match.group("stem") + ".zip"
                        )
                        if not level1_data_path.exists():
                            # Or it may be a sinergise directory, with no suffix.
                            level1_data_path = (
                                level1_metadata_path.parent / match.group("stem")
                            )
                            if not level1_data_path.exists():
                                raise ValueError(
                                    f"Can't find level1 data for {level1_data_path}"
                                )

                    s2cloudless_prob_path = (
                        s2cloudless_prob_path
                        or wagl_hdf5.with_name(f"{granule_name}.prob.s2cloudless.tif")
                    )
                    if not s2cloudless_prob_path.exists():
                        raise ValueError(
                            f"No s2cloudless probability image found at {s2cloudless_prob_path}"
                        )

                    s2cloudless_mask_path = (
                        s2cloudless_mask_path
                        or wagl_hdf5.with_name(f"{granule_name}.mask.s2cloudless.tif")
                    )
                    if not s2cloudless_mask_path.exists():
                        raise ValueError(
                            f"No s2cloudless mask image found at {s2cloudless_mask_path}"
                        )

                    s2cloudless_doc_path = s2cloudless_doc_path or wagl_hdf5.with_name(
                        f"{granule_name}.s2cloudless.yaml"
                    )
                    if not s2cloudless_doc_path.exists():
                        raise ValueError(
                            f"No s2cloudless metadata found at {s2cloudless_doc_path}"
                        )
                    with s2cloudless_doc_path.open("r") as fl:
                        [s2cloudless_doc] = loads_yaml(fl)
                else:
                    s2cloudless_prob_path = None
                    s2cloudless_mask_path = None
                    s2cloudless_doc = None

                gqa_doc_path = gqa_doc_path or wagl_hdf5.with_name(
                    f"{granule_name}.gqa.yaml"
                )
                if not gqa_doc_path.exists():
                    raise ValueError(f"No gqa found at {gqa_doc_path}")
                with gqa_doc_path.open("r") as fl:
                    [gqa_doc] = loads_yaml(fl)

                # Optional doc
                if tesp_doc_path:
                    # But if they gave us a path, we're strict about it existing.
                    if not tesp_doc_path.exists():
                        raise ValueError(
                            f"Supplied tesp doc path doesn't exist: {tesp_doc_path}"
                        )
                else:
                    tesp_doc_path = wagl_hdf5.with_name(f"{granule_name}.tesp.yaml")

                tesp_doc = None
                if tesp_doc_path.exists():
                    with tesp_doc_path.open("r") as fl:
                        [tesp_doc] = loads_yaml(fl)

                yield cls(
                    name=granule_name,
                    wagl_hdf5=wagl_hdf5,
                    wagl_metadata=wagl_doc,
                    source_level1_metadata=level1_doc,
                    source_level1_data=level1_data_path,
                    fmask_doc=fmask_doc,
                    fmask_image=fmask_image_path,
                    s2cloudless_prob=s2cloudless_prob_path,
                    s2cloudless_mask=s2cloudless_mask_path,
                    s2cloudless_doc=s2cloudless_doc,
                    gqa_doc=gqa_doc,
                    tesp_doc=tesp_doc,
                )


# Get just  the stem of metadata documents (named 'docname.yaml' or 'docname.odc-metadata.yaml')
METADATA_DOC_STEM = re.compile(
    r"^(?P<stem>.+?)(?:\.odc-metadata)?\.yaml$", re.IGNORECASE
)


def _load_level1_doc(
    level1_data_path: Path,
    user_specified_l1_path: Optional[Path] = None,
    allow_missing_provenance=False,
) -> DatasetDoc:
    if user_specified_l1_path:
        if not user_specified_l1_path.exists():
            raise ValueError(
                f"No level1 metadata found at given path {user_specified_l1_path}"
            )
        level1_path = user_specified_l1_path
    else:
        # We try to infer it from the data path
        level1_path = level1_data_path

    # If a directory, assume "<dirname>.odc-metadata.yaml"
    if level1_path.is_dir():
        metadata_path = level1_path / (level1_path.name + ".odc-metadata.yaml")
    # Otherwise it's a sibling file with ".odc-metadata.yaml" suffix
    else:
        if level1_path.suffix.lower() == ".yaml":
            metadata_path = level1_path
        else:
            metadata_path = level1_path.with_suffix(".odc-metadata.yaml")

    if not metadata_path.exists():
        if not allow_missing_provenance:
            raise ValueError(
                "No level1 found or provided. "
                f"WAGL said it was at path {str(level1_path)!r}. "
                "Which has no metadata doc we can find, and you didn't specify an alternative. "
                f"(allow_missing_provenance={allow_missing_provenance})"
            )
        return None
    return serialise.from_path(metadata_path)


def package_file(
    out_directory: Path,
    hdf_file: Path,
    included_products: Iterable[str] = DEFAULT_PRODUCTS,
    include_oa: bool = True,
) -> Dict[UUID, Path]:
    """
    Simple alternative to package().

    Takes a single HDF5 and infers other paths (gqa etc) via naming conventions.

    Returns a dictionary of the output datasets: Mapping UUID to the their metadata path.
    """

    out = {}
    for granule in Granule.for_path(hdf_file):
        dataset_id, metadata_path = package(
            out_directory,
            granule,
            included_products=included_products,
            include_oa=include_oa,
        )
        out[dataset_id] = metadata_path

    return out


def package(
    out_directory: Path,
    granule: Granule,
    *,
    product_maturity: ProductMaturity = ProductMaturity.stable,
    included_products: Iterable[str] = DEFAULT_PRODUCTS,
    include_oa: bool = True,
    oa_resolution: Optional[Tuple[float, float]] = None,
    contiguity_resolution: Optional[Tuple[float, float]] = None,
) -> Tuple[UUID, Path]:
    """
    Package an L2 product.

    :param include_oa:

    :param out_directory:
        The base directory for output datasets. A DEA-naming-conventions folder hierarchy
        will be created inside this folder.

    :param granule:
        Granule information. You probably want to make one with Granule.from_path()

    :param included_products:
        A list of imagery products to include in the package.
        Defaults to all products.

    :return:
        The dataset UUID and output metadata path
    """
    included_products = tuple(s.lower() for s in included_products)

    with h5py.File(granule.wagl_hdf5, "r") as fid:
        granule_group = fid[granule.name]

        wagl_doc = _read_wagl_metadata(granule_group)

        with DatasetAssembler(
            out_directory.absolute(),
            # WAGL stamps a good, random ID already.
            dataset_id=granule.wagl_metadata.get("id"),
            naming_conventions=(
                "dea_s2"
                if ("sentinel" in wagl_doc["source_datasets"]["platform_id"].lower())
                else "dea"
            ),
        ) as p:
            _apply_wagl_metadata(p, wagl_doc)

            # It's a GA ARD product.
            p.producer = "ga.gov.au"
            p.product_family = "ard"
            p.maturity = _determine_maturity(
                acq_date=p.datetime,
                processed=p.processed,
                wagl_doc=wagl_doc,
            )

            # We don't bother including product maturity if it's stable, for consistency with old datasets.
            # Stable is the assumed default.
            if product_maturity is not ProductMaturity.stable:
                p.product_maturity = product_maturity

            if granule.source_level1_metadata is not None:
                # For historical consistency: we want to use the instrument that the source L1 product
                # came from, not the instruments reported from the WAGL doc.
                #
                # Eg.
                #     Level 1 will say "OLI_TIRS", while wagl doc will say "OLI".
                #     Our current C3 products say "OLI_TIRS" so we need to stay consistent.
                #     (even though WAGL only *used* the OLI bands, it came from an OLI_TIRS observation)
                #
                # So delete our current wagl one, since we're adding a source dataset:
                if p.instrument is not None:
                    del p.properties["eo:instrument"]

                p.add_source_dataset(
                    granule.source_level1_metadata, auto_inherit_properties=True
                )
                # When level 1 is NRT, ARD is always NRT.
                if granule.source_level1_metadata.maturity == "nrt":
                    p.maturity = "nrt"

            org_collection_number = utils.get_collection_number(
                p.platform, p.producer, p.properties.get("landsat:collection_number")
            )

            p.dataset_version = f"{org_collection_number}.2.1"
            p.region_code = _extract_reference_code(p, granule.name)

            _read_gqa_doc(p, granule.gqa_doc)
            _read_fmask_doc(p, granule.fmask_doc)
            if granule.s2cloudless_doc:
                _read_s2cloudless_doc(p, granule.s2cloudless_doc)
            if granule.tesp_doc:
                _take_software_versions(p, granule.tesp_doc)

            _unpack_products(p, included_products, granule_group, granule)

            if include_oa:
                with sub_product("oa", p):
                    with do("Starting OA", heading=True):
                        resolution_groups = {
                            tuple(granule_group[k].attrs["resolution"]): granule_group[
                                k
                            ]
                            for k in granule_group.keys()
                            if k.startswith("RES-GROUP-")
                        }

                        # Use the highest resolution as the ground sample distance.
                        if "eo:gsd" in p.properties:
                            del p.properties["eo:gsd"]
                        p.properties["eo:gsd"] = min(min(resolution_groups.keys()))

                        _unpack_observation_attributes(
                            p,
                            get_oa_resolution_group(
                                resolution_groups, p.platform, oa_resolution
                            ),
                        )

                    infer_datetime_range = p.platform.startswith("landsat")

                    with do("Contiguity", timedelta=infer_datetime_range):
                        # For landsat, we want the "common" band resolution, not panchromatic. Pick lower res.
                        if contiguity_resolution is not None:
                            contiguity_res = contiguity_resolution
                        elif p.platform.startswith("landsat"):
                            contiguity_res = max(resolution_groups.keys())
                        elif p.platform.startswith("sentinel"):
                            contiguity_res = (10.0, 10.0)

                        if contiguity_res not in resolution_groups:
                            raise ValueError(
                                f"No resolution group {contiguity_res} found in {granule.name}."
                                f"Options: {list(resolution_groups.keys())}"
                            )
                        contiguity_res_grp = resolution_groups[contiguity_res]

                        timedelta_data = (
                            contiguity_res_grp["SATELLITE-SOLAR/TIME-DELTA"]
                            if infer_datetime_range
                            else None
                        )
                        _create_contiguity(
                            p,
                            included_products,
                            resolution_yx=tuple(contiguity_res_grp.attrs["resolution"]),
                            timedelta_data=timedelta_data,
                        )

                    if granule.fmask_image:
                        with do(f"Writing fmask from {granule.fmask_image} "):
                            p.write_measurement(
                                "oa:fmask",
                                granule.fmask_image,
                                expand_valid_data=False,
                                overview_resampling=Resampling.mode,
                                # Because of our strange sub-products and filename standards, we want the
                                # 'oa_' prefix to be included in the recorded band metadata,
                                # but not in its filename.
                                # So we manually calculate a filename without the extra prefix.
                                path=p.names.measurement_filename("fmask"),
                            )

                    if granule.s2cloudless_prob:
                        with do(
                            f"Writing s2cloudless probability from {granule.s2cloudless_prob} "
                        ):
                            p.write_measurement(
                                "oa:s2cloudless_prob",
                                granule.s2cloudless_prob,
                                expand_valid_data=False,
                                overview_resampling=Resampling.bilinear,
                                path=p.names.measurement_filename("s2cloudless-prob"),
                            )

                    if granule.s2cloudless_mask:
                        with do(
                            f"Writing s2cloudless mask from {granule.s2cloudless_mask} "
                        ):
                            p.write_measurement(
                                "oa:s2cloudless_mask",
                                granule.s2cloudless_mask,
                                expand_valid_data=False,
                                overview_resampling=Resampling.mode,
                                path=p.names.measurement_filename("s2cloudless-mask"),
                            )

            with do("Finishing package"):
                return p.done()


def _read_gqa_doc(p: DatasetAssembler, doc: Dict):
    _take_software_versions(p, doc)
    p.extend_user_metadata("gqa", doc)

    # TODO: more of the GQA fields?
    for k, v in flatten_dict(doc["residual"], separator="_"):
        p.properties[f"gqa:{k}"] = v


def _read_fmask_doc(p: DatasetAssembler, doc: Dict):
    for name, value in doc["percent_class_distribution"].items():
        # From Josh: fmask cloud cover trumps the L1 cloud cover.
        if name == "cloud" and not isnan(value):
            if "eo:cloud_cover" in p.properties:
                del p.properties["eo:cloud_cover"]
            p.properties["eo:cloud_cover"] = value

        p.properties[f"fmask:{name}"] = value

    _take_software_versions(p, doc)
    p.extend_user_metadata("fmask", doc)


def _read_s2cloudless_doc(p: DatasetAssembler, doc: Dict):
    for name, value in doc["percent_class_distribution"].items():
        p.properties[f"s2cloudless:{name}"] = value

    _take_software_versions(p, doc)
    p.extend_user_metadata("s2cloudless", doc)


def _take_software_versions(p: DatasetAssembler, doc: Dict):
    versions = doc.pop("software_versions", {})

    for name, o in versions.items():
        p.note_software_version(name, o.get("repo_url"), o.get("version"))


def find_a_granule_name(wagl_hdf5: Path) -> str:
    """
    Try to extract granule name from wagl filename,

    >>> find_a_granule_name(Path('LT50910841993188ASA00.wagl.h5'))
    'LT50910841993188ASA00'
    >>> find_a_granule_name(Path('S2A_OPER_MSI_L1C_TL_EPAE_20201031T022859_A027984_T53JQJ_N02.09.wagl.h5'))
    'S2A_OPER_MSI_L1C_TL_EPAE_20201031T022859_A027984_T53JQJ_N02.09'
    >>> find_a_granule_name(Path('my-test-granule.h5'))
    Traceback (most recent call last):
    ...
    ValueError: Does not appear to be a wagl output filename? 'my-test-granule.h5'.
    """
    if not wagl_hdf5.name.endswith(".wagl.h5"):
        raise ValueError(
            f"Does not appear to be a wagl output filename? {wagl_hdf5.name!r}."
        )

    return wagl_hdf5.name[: -len(".wagl.h5")]


def _read_wagl_metadata(granule_group: h5py.Group):
    try:
        wagl_path, *ancil_paths = (
            pth for pth in _find_h5_paths(granule_group, "SCALAR") if "METADATA" in pth
        )
    except ValueError:
        raise ValueError("No nbar metadata found in granule")

    [wagl_doc] = loads_yaml(granule_group[wagl_path][()])

    for i, path in enumerate(ancil_paths, start=2):
        wagl_doc.setdefault(f"wagl_{i}", {}).update(
            next(iter(loads_yaml(granule_group[path][()])))["ancillary"]
        )
    return wagl_doc


def _apply_wagl_metadata(p: DatasetAssembler, wagl_doc: Dict):
    source = wagl_doc["source_datasets"]

    p.datetime = source["acquisition_datetime"]
    p.platform = source["platform_id"]
    p.instrument = source["sensor_id"]

    if "final_ancillaries" in wagl_doc:
        p.properties["dea:final_ancillaries"] = wagl_doc["final_ancillaries"]

    try:
        p.processed = get_path(wagl_doc, ("system_information", "time_processed"))
    except PathAccessError:
        raise RuntimeError("WAGL dataset contains no processed time.")

    _take_software_versions(p, wagl_doc)
    p.extend_user_metadata("wagl", wagl_doc)


def _determine_maturity(acq_date: datetime, processed: datetime, wagl_doc: Dict):
    """
    Determine maturity field of a dataset.

    Based on the fallback logic in nbar pages of CMI, eg: https://cmi.ga.gov.au/ga_ls5t_nbart_3
    """

    ancillary_tiers = {
        key.lower(): o["tier"]
        for key, o in wagl_doc["ancillary"].items()
        if "tier" in o
    }

    if "water_vapour" not in ancillary_tiers:
        # Perhaps this should be a warning, but I'm being strict until told otherwise.
        # (a warning is easy to ignore)
        raise ValueError(
            f"No water vapour ancillary tier. Got {list(ancillary_tiers.keys())!r}"
        )

    water_vapour_is_definitive = ancillary_tiers["water_vapour"].lower() == "definitive"

    if (processed - acq_date) < timedelta(hours=48):
        return "nrt"

    if not water_vapour_is_definitive:
        return "interim"

    # For accurate BRDF, both Aqua and Terra need to be operating.
    # Aqua launched May 2002, and we add a ~2 month buffer of operation.
    if acq_date < default_utc(datetime(2002, 7, 1)):
        return "final"

    if "brdf" not in ancillary_tiers:
        # Perhaps this should be a warning, but I'm being strict until told otherwise.
        # (a warning is easy to ignore)
        raise ValueError(
            f"No brdf tier available. Got {list(ancillary_tiers.keys())!r}"
        )
    brdf_tier = ancillary_tiers["brdf"].lower()

    if "definitive" in brdf_tier:
        return "final"
    elif "fallback" in brdf_tier:
        return "interim"
    else:
        # This value should not occur for production data, only for experiments
        return "user"
